import logging

import grpc

import departure.board.view_model as view_model
import departure.board.contents as contents
import departure.commons.helpers as helpers

import departure.board.departure_pb2 as departure_pb2
import departure.board.departure_pb2_grpc as departure_pb2_grpc  # for type hinting
import departure.board.protobuf as protobuf

logger = logging.getLogger(__name__)


# pylint: disable=abstract-method
class ViewModelTflTube(view_model.ViewModel):
    pass


# pylint: disable=abstract-method
class ViewModelTflTube_192_32(ViewModelTflTube):
    def __init__(self):
        self.board_text = contents.BoardText(
            font=contents.read_bdf_font("fonts/6x10_proportional.bdf"),
            colour_map={
                "orange": (255, 204, 0),
                "red": (255, 0, 0),
                "green": (0, 255, 0),
                "white": (255, 255, 255),
            },
            tabs=[[0, "l"], [10, "l"], [191, "r"]],
        )

    def pixels_single_line(self, train, pos: str):
        return self.board_text.colour_text_tabbed_row_pixels(
            [
                [pos, "orange"],
                [train["towards"], "orange"],
                [helpers.arrival_time_en(train["time_to_station"]), "orange"],
            ]
        )

    def next_content(self, next_trains):
        # empty board if called with None
        if next_trains is None:
            return [], (0, 0), [], (0, 0), [], (0, 0)

        # sort trains by time to station
        next_trains = sorted(next_trains.values(), key=lambda a: a["time_to_station"])

        # top section
        if len(next_trains) == 0:  # "No service" if no service
            top_section_content_pixels = []
            top_section_content_pixels_size = (0, 0)
        else:  # otherwise: 1st train
            (
                top_section_content_pixels,
                top_section_content_pixels_size,
            ) = self.pixels_single_line(next_trains[0], "1")

        # middle section
        if len(next_trains) < 2:  # nothing if fewer than 2 trains
            middle_section_content_pixels = []
            middle_section_content_pixels_size = (0, 0)
        else:  # otherwise: 2nd train
            (
                middle_section_content_pixels,
                middle_section_content_pixels_size,
            ) = self.pixels_single_line(next_trains[1], "2")

        # bottom section
        if len(next_trains) < 3:  # nothing if fewer than 3 trains
            bottom_section_content_pixels = []
            bottom_section_content_pixels_size = (0, 0)

        elif len(next_trains) == 3:  # 3rd train (if only 3 trains)
            (
                bottom_section_content_pixels,
                bottom_section_content_pixels_size,
            ) = self.pixels_single_line(next_trains[1], "3")

        else:  # 3rd-4th trains (if more than 3 trains)
            (
                bottom_section_content_pixels,
                bottom_section_content_pixels_size,
            ) = self.board_text.colour_text_tabbed_rows_pixels(
                [
                    [
                        ["3", "orange"],
                        [next_trains[2]["towards"], "orange"],
                        [
                            helpers.arrival_time_en(next_trains[2]["time_to_station"]),
                            "orange",
                        ],
                    ],
                    [
                        ["4", "orange"],
                        [next_trains[3]["towards"], "orange"],
                        [
                            helpers.arrival_time_en(next_trains[3]["time_to_station"]),
                            "orange",
                        ],
                    ],
                ],
                row_height=11,
            )

        return (
            top_section_content_pixels,
            top_section_content_pixels_size,
            middle_section_content_pixels,
            middle_section_content_pixels_size,
            bottom_section_content_pixels,
            bottom_section_content_pixels_size,
        )


class ViewModelTflTube_192_32_3_Rows_To_ProtocolBuffers(ViewModelTflTube_192_32):
    def __init__(self, board_manager_stub: departure_pb2_grpc.BoardManagerStub):
        super().__init__()
        self.board_manager_stub = board_manager_stub

        self.next_service_tracker = [
            departure_pb2.Movement(
                static_content=departure_pb2.StaticContent(total_duration=2000)
            ),
            departure_pb2.Movement(
                scrolling_content=departure_pb2.ScrollingContent(
                    step_duration=100,
                    total_steps=11,
                    delta_x_per_step=0,
                    delta_y_per_step=-1,
                )
            ),
        ]

        self.no_movement = [
            departure_pb2.Movement(no_movement=departure_pb2.NoMovement())
        ]

    def update(self, next_trains):  # pylint: disable=arguments-differ
        requests_data = []

        (
            top_section_content_pixels,
            top_section_content_pixels_size,
            middle_section_content_pixels,
            middle_section_content_pixels_size,
            bottom_section_content_pixels,
            bottom_section_content_pixels_size,
        ) = self.next_content(next_trains)

        # top section: 1st train
        requests_data.append(
            departure_pb2.BoardSectionUpdateRequest(
                section_index=0,
                content=departure_pb2.BoardSectionContent(
                    pixels=protobuf.serialise_pixels(top_section_content_pixels),
                    content_w=top_section_content_pixels_size[0],
                    content_h=top_section_content_pixels_size[1],
                    repeat_x=False,
                    repeat_y=False,
                ),
                movement=self.no_movement,
                continue_movement=False,  # force reset
            )
        )

        # middle section: 2nd train
        requests_data.append(
            departure_pb2.BoardSectionUpdateRequest(
                section_index=1,
                content=departure_pb2.BoardSectionContent(
                    pixels=protobuf.serialise_pixels(middle_section_content_pixels),
                    content_w=middle_section_content_pixels_size[0],
                    content_h=middle_section_content_pixels_size[1],
                    repeat_x=False,
                    repeat_y=False,
                ),
                movement=self.no_movement,
                continue_movement=False,  # force reset
            )
        )

        # bottom section: 3rd train onwards
        if next_trains is None or len(next_trains) <= 3:
            next_movement = self.no_movement
        else:
            next_movement = self.next_service_tracker

        requests_data.append(
            departure_pb2.BoardSectionUpdateRequest(
                section_index=2,
                content=departure_pb2.BoardSectionContent(
                    pixels=protobuf.serialise_pixels(bottom_section_content_pixels),
                    content_w=bottom_section_content_pixels_size[0],
                    content_h=bottom_section_content_pixels_size[1],
                    repeat_x=False,
                    repeat_y=True,
                ),
                movement=next_movement,
                continue_movement=True,
            )
        )

        # send gRPC message
        try:
            self.board_manager_stub.BoardSectionsUpdate(
                departure_pb2.BoardSectionsUpdateRequest(requests=requests_data)
            )
        # pylint: disable=protected-access
        except (grpc._channel._Rendezvous, grpc._channel._InactiveRpcError):
            logger.warning("connection to board failed")



# 1) set provider environment variables
# 2) set IP address of board in environment variable DEPARTURE_BOARD_SERVER
# 3) uvicorn api_server:app --reload --host 0.0.0.0

import time
import os

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from starlette.responses import FileResponse

import departure.provider.national_rail.server as national_rail_server
import departure.provider.ratp.server as ratp_server
import departure.provider.sncf.server as sncf_server
import departure.provider.tfl_tube.server as tfl_tube_server
import departure.provider.transilien.server as transilien_server
import departure.provider.ns.server as ns_server
from departure.board import board_client
from departure.commons.log import init_logging
from . import admin


# initialise logging
init_logging()

# initialise app
app = FastAPI()

# initialise shared board client
app.board_client = board_client.BoardClient()

### API routes first
app.include_router(national_rail_server.router, prefix="/national-rail")
app.include_router(ratp_server.router, prefix="/ratp")
app.include_router(sncf_server.router, prefix="/sncf")
app.include_router(tfl_tube_server.router, prefix="/tfl-tube")
app.include_router(transilien_server.router, prefix="/transilien")
app.include_router(ns_server.router, prefix="/ns")


@app.post("/stop-client")
async def stop_client():
    # stop board client if running
    if app.board_client.running:
        app.board_client.running = False
        time.sleep(1)

    return {"status": "OK"}


@app.get("/client-status")
async def client_status():
    if app.board_client.running:
        return {"client_status": "running"}
    return {"client_status": "stopped"}


@app.get("/board-server-status")
async def board_server_status():
    return admin.is_board_server_up()


@app.post("/shutdown-board-server")
async def shutdown_board_server():
    await stop_client()
    return admin.shutdown_board_server()


@app.post("/shutdown-web-server")
async def shutdown_web_server():
    admin.shutdown_web_server()
    return {"status": "OK"}


### static files second (https://github.com/tiangolo/fastapi/issues/130) if root set
if "DEPARTURE_STATIC_WEBROOT" in os.environ:
    # index
    @app.get("/")
    async def read_index():
        return FileResponse(f"{os.environ['DEPARTURE_STATIC_WEBROOT']}/index.html")

    # other static files
    app.mount(
        "/",
        StaticFiles(directory=os.environ["DEPARTURE_STATIC_WEBROOT"]),
        name="static",
    )



import json
import pathlib
import os

import departure.commons.helpers as helpers
from . import api

DATA_DIRNAME = "data"
DATA_PATH = pathlib.Path(__file__).parents[0] / DATA_DIRNAME
STATIONS_DATA_FILENAME = "stations.json"

lines = {
    "bakerloo": {
        "name": "Bakerloo",
        "canonical_directions": {
            "Southbound": "inbound",
            "Northbound": "outbound",
        },
        "termini": {  # Bakerloo Line WTT 43 (10-Dec-2017)
            "940GZZLUEAC": {
                "is_inline": False,
                "arrival": "Southbound",
                "departure": "Northbound",
            },
            "940GZZLUHAW": {
                "is_inline": False,
                "arrival": "Northbound",
                "departure": "Southbound",
            },
            "940GZZLUQPS": {
                "is_inline": True,
                "arrival": "Northbound",
                "departure": "Southbound",
            },
            "940GZZLUSGP": {
                "is_inline": True,
                "arrival": "Northbound",
                "departure": "Southbound",
            },
        },
    },
    "central": {
        "name": "Central",
        # All trains enter the loop via the Inner Rail (from LYS Eastbound /
        # outbound) and leave it via the Outer Rail (at LYS Westbound / inbound),
        # and the loop ends at HLT or WOF (doesn't actually loop around).
        # Therefore, for our purposes, the Inner Rail (resp. Outer Rail) behaves
        # like an extension of the Eastbound (reps. Westbound) services.
        # In reality, canonical directions match geographical directions,
        # i.e.:
        # - inbound is Westbound => Inner Rail north of HLT, Outer Rail south of
        #     HLT
        # - outbound is Eastbound => Outer Rail north of HLT, Inner Rail south
        #     of HLT
        # > see hainault_loop_outer_rail_outbound_stations
        "canonical_directions": {
            "Westbound": "inbound",
            "Eastbound": "outbound",
            "Outer Rail": "inbound",  # actually opposite north of HLT
            "Inner Rail": "outbound",  # actually opposite north of HLT
        },
        "termini": {  # Central Line WTT 69 (07-Aug-2016)
            "940GZZLUNAN": {
                "is_inline": True,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUEBY": {
                "is_inline": False,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUEPG": {
                "is_inline": False,
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
            "940GZZLUHLT": {
                "is_inline": True,  # only when coming from south
                "arrival": "Inner Rail",
                "departure": "Outer Rail",
            },
            "940GZZLUNBP": {
                "is_inline": True,  # only when coming from south
                "arrival": "Inner Rail",
                "departure": "Outer Rail",
            },
            "940GZZLUWOF": {
                "is_inline": True,
                "arrival": "Inner Rail",  # Westbound - Platform 2
                "departure": "Outer Rail",  # Eastbound - Platform 3
            },
            "940GZZLULGN": {
                "is_inline": True,
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
            "940GZZLUNHT": {
                "is_inline": True,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUWRP": {
                "is_inline": False,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUWCY": {
                "is_inline": True,
                "arrival": "Westbound",
                "departure": "Eastbound",
                "stations_inverse_terminating_direction": [
                    "940GZZLUEBY",
                    "940GZZLUWTA",
                    "940GZZLUNAN",
                    "940GZZLUEAN",
                ],
            },
        },
    },
    "circle": {
        "name": "Circle",
        "canonical_directions": {
            "Westbound": None,
            "Eastbound": None,
        },
        "merged_lines": ["district", "hammersmith-city"],
        "termini": {
            "940GZZLUHSC": {
                "is_inline": False,
                "arrival": "Westbound",
                "departure": "Eastbound",
                "arrival_rail": "Inner Rail",
                "departure_rail": "Outer Rail",
            },
            "940GZZLUERC": {  # HSC-ALD-ERC, marked as "Edgware Road (Circle)"
                "is_inline": True,
                "arrival": "Eastbound",
                "departure": "Westbound",
                "arrival_rail": "Outer Rail",
                "departure_rail": "Inner Rail",
            },
        },
    },
    "district": {
        "name": "District",
        "canonical_directions": {
            "Westbound": "inbound",
            "Eastbound": "outbound",
        },
        "merged_lines": ["circle", "hammersmith-city"],
        "termini": {  # District Line WTT 150 (20-May-2018)
            "940GZZLUBKG": {
                "is_inline": True,
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
            "940GZZLUEBY": {
                "is_inline": False,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUERC": {
                "is_inline": False,
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
            "940GZZLURMD": {
                "is_inline": False,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUTWH": {
                "is_inline": True,  # WIM-TWH
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
            "940GZZLUKOY": {
                "is_inline": True,  # HSK-KOY
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUUPM": {
                "is_inline": False,
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
            "940GZZLUWIM": {
                "is_inline": False,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
        },
    },
    "hammersmith-city": {
        "name": "Hammersmith & City",
        "canonical_directions": {
            "Westbound": "inbound",
            "Eastbound": "outbound",
        },
        "merged_lines": ["circle", "district"],
        "termini": {  # Circle and H&C Line WTT 35 (21-May-2017)
            "940GZZLUERC": {
                "is_inline": True,
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
            "940GZZLUBKG": {
                "is_inline": False,
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
            "940GZZLUHSC": {
                "is_inline": False,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
        },
    },
    "jubilee": {
        "name": "Jubilee",
        "canonical_directions": {
            "Westbound": "inbound",
            "Northbound": "inbound",
            "Eastbound": "outbound",
            "Southbound": "outbound",
        },
        "termini": {  # Jubilee Line WTT 15 (20-May-2018)
            "940GZZLUSTM": {
                "is_inline": False,
                "arrival": "Northbound",
                "departure": "Southbound",
            },
            "940GZZLUSTD": {
                "is_inline": False,
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
            "940GZZLUWYP": {
                "is_inline": True,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
        },
    },
    "metropolitan": {
        "name": "Metropolitan",
        "canonical_directions": {
            "Westbound": "inbound",
            "Northbound": "inbound",
            "Eastbound": "outbound",
            "Southbound": "outbound",
        },
        "termini": {  # Metropolitan Line WTT 341 (30-Dec-2018)
            "940GZZLUCSM": {
                "is_inline": False,
                "arrival": "Northbound",
                "departure": "Southbound",
            },
            "940GZZLUALD": {
                "is_inline": False,
                "arrival": "Southbound",
                "departure": "Northbound",
            },
            "940GZZLUAMS": {
                "is_inline": False,
                "arrival": "Northbound",
                "departure": "Southbound",
            },
            "940GZZLUUXB": {
                "is_inline": False,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUWAF": {
                "is_inline": False,
                "arrival": "Northbound",
                "departure": "Southbound",
            },
            "940GZZLUBST": {
                "is_inline": True,  # WAF-BST
                "arrival": "Southbound",
                "departure": "Northbound",
            },
        },
    },
    "northern": {  # Northern Line WTT 57 (29-Jan-2018)
        "name": "Northern",
        "canonical_directions": {
            "Southbound": "inbound",
            "Northbound": "outbound",
        },
        "termini": {
            "940GZZLUEGW": {
                "is_inline": False,
                "arrival": "Northbound",
                "departure": "Southbound",
            },
            "940GZZLUHBT": {
                "is_inline": False,
                "arrival": "Northbound",
                "departure": "Southbound",
            },
            "940GZZLUKNG": {
                "is_inline": True,
                "arrival": "Southbound",
                "departure": "Northbound",
            },
            "940GZZLUMHL": {
                "is_inline": False,
                "arrival": "Northbound",
                "departure": "Southbound",
            },
            "940GZZLUMDN": {
                "is_inline": False,
                "arrival": "Southbound",
                "departure": "Northbound",
            },
        },
    },
    "piccadilly": {
        "name": "Piccadilly",
        "canonical_directions": {
            "Westbound": "inbound",
            "Eastbound": "outbound",
        },
        "termini": {
            "940GZZLUHR4": {
                "is_inline": True,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUHR5": {
                "is_inline": False,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUHRC": {
                "is_inline": True,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUASG": {
                "is_inline": True,
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
            "940GZZLUCKS": {
                "is_inline": False,
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
            "940GZZLUNFD": {
                "is_inline": True,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUUXB": {
                "is_inline": False,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
            "940GZZLUWOG": {
                "is_inline": True,
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
        },
    },
    "victoria": {
        "name": "Victoria",
        "canonical_directions": {
            "Southbound": "inbound",
            "Northbound": "outbound",
        },
        "termini": {
            "940GZZLUWWL": {
                "is_inline": False,
                "arrival": "Northbound",
                "departure": "Southbound",
            },
            "940GZZLUBXN": {
                "is_inline": False,
                "arrival": "Southbound",
                "departure": "Northbound",
            },
        },
    },
    "waterloo-city": {
        "name": "Waterloo & City",
        "canonical_directions": {
            "Westbound": "inbound",
            "Eastbound": "outbound",
        },
        "termini": {
            "940GZZLUBNK": {
                "is_inline": False,
                "arrival": "Eastbound",
                "departure": "Westbound",
            },
            "940GZZLUWLO": {
                "is_inline": False,
                "arrival": "Westbound",
                "departure": "Eastbound",
            },
        },
    },
}


# ordered list of stations on Circle Line on the outer rail
circle_line_outer_rail_stations = [  # reserved for possible future use
    "940GZZLUHSC",
    "940GZZLUGHK",
    "940GZZLUSBM",
    "940GZZLUWLA",
    "940GZZLULRD",
    "940GZZLULAD",
    "940GZZLUWSP",
    "940GZZLURYO",
    "940GZZLUPAC",
    "940GZZLUERC",  # through ERC
    "940GZZLUBST",
    "940GZZLUGPS",
    "940GZZLUESQ",
    "940GZZLUKSX",
    "940GZZLUFCN",
    "940GZZLUBBN",
    "940GZZLUMGT",
    "940GZZLULVT",
    "940GZZLUALD",
    "940GZZLUTWH",
    "940GZZLUMMT",
    "940GZZLUCST",
    "940GZZLUMSH",
    "940GZZLUBKF",
    "940GZZLUTMP",
    "940GZZLUEMB",
    "940GZZLUWSM",
    "940GZZLUSJP",
    "940GZZLUVIC",
    "940GZZLUSSQ",
    "940GZZLUSKT",
    "940GZZLUGTR",
    "940GZZLUHSK",
    "940GZZLUNHG",
    "940GZZLUBWT",
    "940GZZLUPAC",
    "940GZZLUERC",  # terminus ERC
]


# list of stations on Circle Line that are eastbound on the outer rail
circle_line_outer_rail_eastbound_stations = [
    "940GZZLUHSC",
    "940GZZLUGHK",
    "940GZZLUSBM",
    "940GZZLUWLA",
    "940GZZLULRD",
    "940GZZLULAD",
    "940GZZLUWSP",
    "940GZZLURYO",
    "940GZZLUPAC",
    "940GZZLUERC",
    "940GZZLUBST",
    "940GZZLUGPS",
    "940GZZLUESQ",
    "940GZZLUKSX",
    "940GZZLUFCN",
    "940GZZLUBBN",
    "940GZZLUMGT",
    "940GZZLULVT",
    "940GZZLUALD",
    "940GZZLUHSK",
    "940GZZLUNHG",
    "940GZZLUBWT",
    "940GZZLUPAC",
]

# list of stations on Circle Line that are westbound on the outer rail
circle_line_outer_rail_westbound_stations = [
    "940GZZLUTWH",
    "940GZZLUMMT",
    "940GZZLUCST",
    "940GZZLUMSH",
    "940GZZLUBKF",
    "940GZZLUTMP",
    "940GZZLUEMB",
    "940GZZLUWSM",
    "940GZZLUSJP",
    "940GZZLUVIC",
    "940GZZLUSSQ",
    "940GZZLUSKT",
    "940GZZLUGTR",
]


alternative_station_names = {  # in currentLocation or towards values
    "Shepherd's Bush": "940GZZLUSBC",
    "Kings Cross": "940GZZLUKSX",
    "King's Cross": "940GZZLUKSX",
    "Kings Cross St. Pancras": "940GZZLUKSX",
    "Elephant and Castle": "940GZZLUEAC",
    "Harrow and Wealdstone": "940GZZLUHAW",
    "Harrow-On-The-Hill": "940GZZLUHOH",
    "Edgware Road (Circle)": "940GZZLUERC",
    "Regents Park": "940GZZLURGP",
    "Heathrow Terminal 1,2,3": "940GZZLUHRC",
    "Chalfont and Latimer": "940GZZLUCAL",
}


alternative_terminus_names = {  # in towards values
    "Heathrow T123 + 5": "940GZZLUHR5",
    "Heathrow via T4 Loop": "940GZZLUHR4",
}


alternative_station_names_ambiguous = {
    "Edgware Road": {
        "bakerloo": "940GZZLUERB",
        "circle": "940GZZLUERC",
        "district": "940GZZLUERC",
        "hammersmith-city": "940GZZLUERC",
    },
    "Hammersmith": {
        "district": "940GZZLUHSD",
        "piccadilly": "940GZZLUHSD",
        "circle": "940GZZLUHSC",
        "hammersmith-city": "940GZZLUHSC",
    },
}


non_station_locations = [
    "Newbury Park Loop",
    "North Acton Junction",
    "Leaving Stanmore",
    "Wood Green Sidings",
    "Northolt Sidings",
]


platform_name_prefix_direction = {  # tight matching
    "Westbound - Platform ": "Westbound",
    "Eastbound - Platform ": "Eastbound",
    "Northbound - Platform ": "Northbound",
    "Southbound - Platform ": "Southbound",
    "Northbound Fast - Platform ": "Northbound",
    "Southbound Fast - Platform ": "Southbound",
    "Outer Rail - Platform ": "Outer Rail",
    "Inner Rail - Platform ": "Inner Rail",
    "Westbound Platform ": "Westbound",
    "WestBound - Platform ": "Westbound",
    "EastBound - Platform ": "Eastbound",
}


hainault_loop_outer_rail_outbound_stations = [
    "940GZZLURVY",
    "940GZZLUCWL",
    "940GZZLUGGH",
]


hainault_loop_real_canonical_directions = [  # reserved for possible future use
    # match outbound = Eastbound, inbound = Westbound
    # north of HLT, Outer Rail is Eastbound
    {
        "stations": ["940GZZLURVY", "940GZZLUCWL", "940GZZLUGGH"],
        "Outer Rail": "outbound",
        "Inner Rail": "inbound",
    },
    # HLT => depends if arriving/departing
    {
        "stations": ["940GZZLUHLT"],
        "Outer Rail": {
            "arrival": "outbound",
            "departure": "inbound",
        },
        "Inner Rail": {
            "arrival": "inbound",
            "departure": "outbound",
        },
    },
    # south of HLT, Outer Rail is Westbound
    {
        "stations": [
            "940GZZLUFLP",
            "940GZZLUBKE",
            "940GZZLUNBP",
            "940GZZLUGTH",
            "940GZZLURBG",
            "940GZZLUWSD",
        ],
        "Outer Rail": "inbound",
        "Inner Rail": "outbound",
    },
]


def stoppoints_by_line(line_id: str):
    line_data_file_name = DATA_PATH / (line_id + ".json")

    if os.path.exists(line_data_file_name):
        # read data from file if it exists
        with open(line_data_file_name, "r", encoding="utf-8") as line_data_file:
            stoppoints = json.load(line_data_file)

    else:
        # download missing line data otherwise
        stoppoints = api.line_stoppoints(line_id)
        with open(line_data_file_name, "w") as line_data_file:
            json.dump(stoppoints, line_data_file)

    return stoppoints


def add_line_data_to_stations_dict(stations_dict: dict, line_id: str):
    stoppoints = stoppoints_by_line(line_id)

    # process stop point data
    for stoppoint in stoppoints:
        # new station
        if stoppoint["naptanId"] not in stations_dict:
            stations_dict[stoppoint["naptanId"]] = {
                "commonName": stoppoint["commonName"],
                "lines": [],
            }

            # friendly name
            if stoppoint["commonName"].endswith(" Underground Station"):
                stations_dict[stoppoint["naptanId"]]["name"] = helpers.rchop(
                    stoppoint["commonName"], " Underground Station"
                )
            elif stoppoint["commonName"].endswith("-Underground"):
                # for "Paddington (H&C Line)-Underground" (PAH)
                stations_dict[stoppoint["naptanId"]]["name"] = helpers.rchop(
                    stoppoint["commonName"], "-Underground"
                )

        # add current line
        stations_dict[stoppoint["naptanId"]]["lines"].append(line_id)

    return stations_dict


def _stations(data_path, stations_data_filename):
    # read from stations data file if available
    stations_data_file_path = data_path / stations_data_filename
    if stations_data_file_path.exists():
        with stations_data_file_path.open(mode="r", encoding="utf-8") as stations_file:
            stations = json.load(stations_file)
        return stations

    # otherwise generate data from individual line files
    stations = {}
    for line_id in lines:
        stations = add_line_data_to_stations_dict(_stations, line_id)

    # read data from file
    with open(stations_data_file_path, "w") as stations_file:
        json.dump(_stations, stations_file)

    return stations


def _station_id_by_name(stations):
    return {stations[station_id]["name"]: station_id for station_id in stations}


STATIONS = _stations(DATA_PATH, STATIONS_DATA_FILENAME)
STATION_ID_BY_NAME = _station_id_by_name(STATIONS)



import time
import threading
import logging

from departure.board.board import Board
from departure.board.renderer import Renderer

logger = logging.getLogger(__name__)


class BoardAnimator(threading.Thread):
    def __init__(
        self,
        board: Board,
        renderer: Renderer,
        time_step_in_s: float,
        board_lock: threading.RLock,
        end_event: threading.Event,
    ):
        super().__init__()
        self.board = board
        self.renderer = renderer
        self.time_step_in_s = time_step_in_s
        self.board_lock = board_lock
        self.end_event = end_event

    def update(self, delta_time: int):
        self.board.update(delta_time)
        self.renderer.render_frame(self.board.pixels())

    def run(self):
        self.board.reset()

        # start the clock
        last_loop_start_time = time.monotonic()
        next_loop_start_time = time.monotonic() + self.time_step_in_s

        # code for non-drifting loop from
        # https://raspberrypi.stackexchange.com/a/34245

        while not self.end_event.is_set():
            this_loop_start_time = time.monotonic()

            # pylint: disable=pointless-string-statement
            """
            # display drift (debugging)
            logger.debug('drift {:.02f} ms'.format(
                1000 * (this_loop_start_time - next_loop_start_time
                    + self.time_step_in_s)
                ))
            """

            with self.board_lock:
                self.update(int(1000 * (time.monotonic() - last_loop_start_time)))

            delay = -self.time_step_in_s
            missed = 0

            while delay < 0.0:
                delay = next_loop_start_time - time.monotonic()
                next_loop_start_time += self.time_step_in_s
                missed += 1

            if missed > 1:
                logger.warning(
                    "exceeded animation loop step duration %s time(s)", missed - 1
                )

            self.end_event.wait(delay)
            last_loop_start_time = this_loop_start_time

        logger.info("board animator stopped running")



import logging

import grpc

import departure.board.view_model as view_model
import departure.board.contents as contents
import departure.commons.helpers as helpers
import departure.board.departure_pb2 as departure_pb2
import departure.board.protobuf as protobuf

# type hinting
import departure.board.departure_pb2_grpc as departure_pb2_grpc

from . import data

logger = logging.getLogger(__name__)


def text_array_train_overview(train, pos: int):
    if train["actual_time"] == train["planned_time"]:
        time_colour = "green"
    else:
        time_colour = "red"

    return [
        [helpers.ordinal_en(pos), "orange"],
        [train["actual_time"], time_colour],
        [train["track"], "white"],
        [train["direction"], "orange", 106],
        [train["train_number"], "white"],
    ]


# for type hinting
# pylint: disable=abstract-method,too-few-public-methods
class ViewModelNS(view_model.ViewModel):
    pass


class ViewModelNS192x32x3(ViewModelNS):  # pylint: disable=abstract-method
    def __init__(self):
        self.board_text = contents.BoardText(
            font=contents.read_bdf_font("fonts/6x10_condensed.bdf"),
            colour_map={
                "orange": (255, 204, 0),
                "red": (255, 0, 0),
                "green": (0, 255, 0),
                "white": (255, 255, 255),
            },
            tabs=[[0, "l"], [20, "l"], [58, "r"], [62, "l"], [191, "r"]],
        )

    def pixels_train_overview(self, train, pos: int):
        return self.board_text.colour_text_tabbed_row_pixels(
            text_array_train_overview(train, pos)
        )

    def pixels_train_details(self, train: dict, current_station_ns_code: str):
        stops = []
        index_last_station = len(train["schedule"]) - 1

        stations_by_uic = data.STATIONS_BY_UIC
        current_station_uic = data.STATIONS_BY_NS_CODE[current_station_ns_code]["uic"]
        before_current_station = True
        index_after_current_station = 0

        for i, stop in enumerate(train["schedule"]):
            # ignore stops before current stop
            if stop["stop_uic"] == current_station_uic:
                before_current_station = False
                index_after_current_station = i + 1
                continue

            if before_current_station:
                continue

            # display schedule after current station

            ## list formatting (e.g. Oxford comma)
            if i == index_after_current_station:
                pass
            elif i == index_last_station:
                stops.append([", and ", "orange"])
            else:
                stops.append([", ", "orange"])

            ## resolve name
            if stop["stop_uic"] in stations_by_uic:
                stop_name = stations_by_uic[stop["stop_uic"]]["name_medium"]
            else:  # don't attempt to resolve using UIC if stop is outside NL
                stop_name = f"{stop['stop_name']}"
            stops.append([f"{stop_name} ", "orange"])

            if stop["actual_time"] == stop["planned_time"]:
                time_colour = "green"
            else:
                time_colour = "red"

            stops.append([f'({stop["actual_time"]})', time_colour])

        return self.board_text.colour_text_pixels(
            [["Calling at: ", "orange"]]
            + stops
            + [[" (" + train["train_category"] + ")", "orange"]]
        )

    def next_content(self, trains: list, current_station_ns_code: str):
        # empty board if called with None
        if trains is None:
            return [], (0, 0), [], (0, 0), [], (0, 0)

        # top section
        if len(trains) == 0:  # "No service" if no service
            (
                top_section_content_pixels,
                top_section_content_pixels_size,
            ) = self.board_text.colour_text_pixels([["No service", "orange"]])
        else:  # otherwise: 1st train
            (
                top_section_content_pixels,
                top_section_content_pixels_size,
            ) = self.pixels_train_overview(trains[0], 1)

        # middle section
        if len(trains) == 0:  # nothing if no service
            middle_section_content_pixels = []
            middle_section_content_pixels_size = (0, 0)
        else:  # otherwise: calling points of 1st train
            (
                middle_section_content_pixels,
                middle_section_content_pixels_size,
            ) = self.pixels_train_details(trains[0], current_station_ns_code)

        # bottom section
        if len(trains) < 2:  # nothing if fewer than 2 trains
            bottom_section_content_pixels = []
            bottom_section_content_pixels_size = (0, 0)

        elif len(trains) == 2:  # 2nd train (if only 2 trains)
            (
                bottom_section_content_pixels,
                bottom_section_content_pixels_size,
            ) = self.pixels_train_overview(trains[1], 2)

        else:  # 2nd train onwards (if more than 2 trains)
            (
                bottom_section_content_pixels,
                bottom_section_content_pixels_size,
            ) = self.board_text.colour_text_tabbed_rows_pixels(
                [
                    text_array_train_overview(train, i + 2)
                    for i, train in enumerate(trains[1:10])
                ],
                row_height=11,
            )

        return (
            top_section_content_pixels,
            top_section_content_pixels_size,
            middle_section_content_pixels,
            middle_section_content_pixels_size,
            bottom_section_content_pixels,
            bottom_section_content_pixels_size,
        )


class ViewModelNS192x32x3ToProtocolBuffers(ViewModelNS192x32x3):
    def __init__(self, board_manager_stub: departure_pb2_grpc.BoardManagerStub):
        super().__init__()
        self.board_manager_stub = board_manager_stub

        self.next_train_tracker = [
            departure_pb2.Movement(
                static_content=departure_pb2.StaticContent(total_duration=2000)
            ),
            departure_pb2.Movement(
                scrolling_content=departure_pb2.ScrollingContent(
                    step_duration=100,
                    total_steps=11,
                    delta_x_per_step=0,
                    delta_y_per_step=-1,
                )
            ),
        ]

        self.no_movement = [
            departure_pb2.Movement(no_movement=departure_pb2.NoMovement())
        ]

    # pylint: disable=arguments-differ
    def update(self, trains: list, current_station_ns_code: str = ""):
        requests_data = []

        (
            top_section_content_pixels,
            top_section_content_pixels_size,
            middle_section_content_pixels,
            middle_section_content_pixels_size,
            bottom_section_content_pixels,
            bottom_section_content_pixels_size,
        ) = self.next_content(trains, current_station_ns_code)

        # top section: 1st train
        requests_data.append(
            departure_pb2.BoardSectionUpdateRequest(
                section_index=0,
                content=departure_pb2.BoardSectionContent(
                    pixels=protobuf.serialise_pixels(top_section_content_pixels),
                    content_w=top_section_content_pixels_size[0],
                    content_h=top_section_content_pixels_size[1],
                    repeat_x=False,
                    repeat_y=False,
                ),
                movement=self.no_movement,
                continue_movement=False,
            )
        )

        # middle section: calling points of 1st train
        requests_data.append(
            departure_pb2.BoardSectionUpdateRequest(
                section_index=1,
                content=departure_pb2.BoardSectionContent(
                    pixels=protobuf.serialise_pixels(middle_section_content_pixels),
                    content_w=middle_section_content_pixels_size[0] + 192,
                    content_h=middle_section_content_pixels_size[1],
                    repeat_x=True,
                    repeat_y=False,
                ),
                movement=[
                    departure_pb2.Movement(
                        scrolling_content=departure_pb2.ScrollingContent(
                            step_duration=50,
                            total_steps=0,
                            delta_x_per_step=-1,
                            delta_y_per_step=0,
                        )
                    )
                ],
                continue_movement=True,
            )
        )

        # bottom section: 2nd train onwards
        if trains is None or len(trains) <= 2:  # no v-scroll if up to 2 trains
            next_movement = self.no_movement
        else:  # v-scroll if more than 2 trains
            next_movement = self.next_train_tracker

        requests_data.append(
            departure_pb2.BoardSectionUpdateRequest(
                section_index=2,
                content=departure_pb2.BoardSectionContent(
                    pixels=protobuf.serialise_pixels(bottom_section_content_pixels),
                    content_w=bottom_section_content_pixels_size[0],
                    content_h=bottom_section_content_pixels_size[1],
                    repeat_x=True,
                    repeat_y=True,
                ),
                movement=next_movement,
                continue_movement=True,
            )
        )

        # send gRPC message
        try:
            self.board_manager_stub.BoardSectionsUpdate(
                departure_pb2.BoardSectionsUpdateRequest(requests=requests_data)
            )
        # pylint: disable=protected-access
        except (grpc._channel._Rendezvous, grpc._channel._InactiveRpcError):
            logger.warning("connection to board failed")



"""
CLI for Nederlandse Spoorwegen
"""

import threading
import time
import logging

import click

from departure.board import board_client

from . import ui, ns, commons, view_model, data_updater

ONE_DAY_IN_SECONDS = 60 * 60 * 24
logger = logging.getLogger(__name__)


@click.group()
def cli():
    """
    Information and departures for Nederlandse Spoorwegen (NL).

    Note - Your NS API key must be assigned to the NS_API_KEY environment variable.
    """


@click.command()
@click.argument("query_string")
def search(query_string):
    """Search for stations containing QUERY_STRING."""
    stations = ns.search(query_string)
    ui.list_stations(stations)


@click.command(name="next")
@click.argument("ns_code")
@click.option(
    "--full",
    "timetable_for_all_trains",
    is_flag=True,
    help="displays timetable for all trains, instead of just first one",
)
def next_departures(ns_code, timetable_for_all_trains):
    """
    Get next departures at station NS_CODE.

    Use search to find the NS_CODE for a station.
    """

    ns_code = ns_code.upper()
    try:
        trains = ns.departures_with_schedule(ns_code, timetable_for_all_trains)
    except commons.NSException as e:
        print(f"error: {str(e)}")
        return

    ui.list_departures_with_schedule(trains, ns_code)


@click.command(name="board")
@click.argument("ns_code")
def start_board(ns_code):
    """Update a departure board with departures at station NS_CODE.

    Use search to find the NS_CODE for a station."""

    ns_code = ns_code.upper()

    try:
        ns.check_params(ns_code)
    except commons.NSException as e:
        logger.error("error: %s", str(e))
        return

    # start board
    board_client_instance = board_client.BoardClient()

    threading.Thread(
        target=board_client_instance.run,
        args=[
            view_model.ViewModelNS192x32x3ToProtocolBuffers,
            data_updater.DataUpdaterNS,
            {
                "station_code": ns_code,
            },
        ],
    ).start()

    # run until interrupt
    try:
        while True:
            time.sleep(ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        logger.info("received interrupt")
        board_client_instance.running = False


cli.add_command(search)
cli.add_command(next_departures)
cli.add_command(start_board)

if __name__ == "__main__":
    cli()



"""
API for Nederlandse Spoorwegen
"""

import json
import os
import logging

import requests

from . import commons

logger = logging.getLogger(__name__)


def api_request(url: str, payload: dict) -> str:
    commons.check_env_vars()

    try:
        response = requests.get(
            url,
            params=payload,
            timeout=15,
            headers={"Ocp-Apim-Subscription-Key": os.environ["NS_API_KEY"]},
        )
    except requests.exceptions.Timeout:
        logger.warning("NS API HTTP request timed out")
        return None
    except Exception as e:  # pylint: disable=broad-except
        logger.warning("NS API HTTP request error: %s", str(e))
        return None

    # process response
    if response.status_code == 200:
        return json.loads(response.content.decode("utf-8"))

    logger.warning("NS API request failed, HTTP response: %s", response.status_code)
    return None


def departures(station_code):
    url = "https://gateway.apiportal.ns.nl/reisinformatie-api/api/v2/departures"
    payload = {"station": station_code, "lang": "en"}
    return api_request(url, payload)


def journey(train_number):
    url = "https://gateway.apiportal.ns.nl/reisinformatie-api/api/v2/journey"
    payload = {"train": train_number, "lang": "en"}
    return api_request(url, payload)



import pathlib

from bdflib import reader


def read_bdf_font(font_filename):
    font_path = pathlib.Path(__file__).parents[0] / font_filename
    with font_path.open("rb") as f:
        return reader.read_bdf(f)


class BoardTextException(Exception):
    pass


class BoardText:
    def __init__(self, font, colour_map=None, tabs=None):
        self.font = font
        self.colour_map = colour_map
        self.tabs = tabs

    def _glyph_pixels(self, glyph, x_offset, y_offset, colour=(255, 255, 255)):
        # pixel coordinate system (x,y) = (0,0) => top left
        # glyph coordinate system (x,y) = (0,0) => bottom left
        # y_offset = (FONT_ASCENT - 1) + line_number * (FONT_HEIGHT + line_spacing)
        # where FONT_HEIGHT = FONT_ASCENT + FONT_DESCENT
        return [
            (x_offset + x, y_offset - y - glyph.bbY, colour)
            for x in range(glyph.bbW)
            for y in range(glyph.bbH)
            if glyph.data[y] & (1 << (glyph.bbW - x - 1))
        ]

    def monochrome_text_pixels(
        self, text, colour=(255, 255, 255), max_width=0  # 0 == no max width
    ):
        pixels = []
        advance = 0
        for c in text:
            glyph = self.font[ord(c)]
            pixels += self._glyph_pixels(
                glyph, advance, self.font[b"FONT_ASCENT"] - 1, colour
            )
            advance += glyph.advance
            if max_width > 0 and advance > max_width:
                break

        return pixels, (advance, self.font[b"FONT_ASCENT"] + self.font[b"FONT_DESCENT"])

    def colour_text_pixels(self, text_array):
        if self.colour_map is None:
            raise BoardTextException("missing colour map")

        x_size, y_size = 0, 0
        output_pixels = []

        for text_part in text_array:
            text, colour = text_part[:2]
            if len(text_part) == 3:
                max_width = text_part[2]
            else:
                max_width = 0
            pixels, size = self.monochrome_text_pixels(
                text, self.colour_map[colour], max_width
            )
            output_pixels += [(x + x_size, y, c) for x, y, c in pixels]
            x_size += size[0]
        y_size = size[1]
        return output_pixels, (x_size, y_size)

    def colour_text_tabbed_row_pixels(self, text_row):
        if self.tabs is None:
            raise BoardTextException("missing tabs")

        output_pixels = []
        x_size = 0
        for i, text_part in enumerate(text_row):
            tab_pos, tab_type = self.tabs[i]
            text, colour = text_part[:2]
            if len(text_part) == 3:
                max_width = text_part[2]
            else:
                max_width = 0
            pixels, size = self.monochrome_text_pixels(
                text, self.colour_map[colour], max_width
            )

            if tab_type == "l":
                output_pixels += [(x + tab_pos, y, c) for x, y, c in pixels]
                # extend width as needed
                x_size = max(x_size, size[0] + tab_pos)

            elif tab_type == "r":
                output_pixels += [
                    (tab_pos + 1 - size[0] + x, y, c) for x, y, c in pixels
                ]
                # extend width as needed
                x_size = max(x_size, tab_pos + 1)

        y_size = size[1]

        return output_pixels, (x_size, y_size)

    def colour_text_tabbed_rows_pixels(self, text_rows, row_height):
        pixels = []
        x_size = 0
        i = 0

        for i, text_row in enumerate(text_rows):
            row_pixels, size = self.colour_text_tabbed_row_pixels(text_row)
            pixels += [(x, y + i * row_height, c) for x, y, c in row_pixels]
            x_size = max(x_size, size[0])

        y_size = (i + 1) * row_height

        return pixels, (x_size, y_size)



from typing import List, Any

from . import data, commons, api


def check_params(station_code: str) -> None:
    stations = data.STATIONS

    if station_code is not None:
        try:
            _ = stations[station_code]
        except Exception:
            raise commons.NationalRailException(
                f"invalid station code {station_code}"
            ) from Exception


def stations_by_string(string):
    string = str(string).lower().strip()
    results = {}
    stations = data.STATIONS

    # iterate over stations
    for station_id in stations:
        # match?
        if string in stations[station_id].lower():
            results[station_id] = stations[station_id]

    return results


def services_from_station_board_with_details(station_board_with_details):
    services = []

    # end here if no services
    if station_board_with_details.trainServices is None:
        return services

    try:
        for station_board_service in station_board_with_details.trainServices.service:
            services.append(
                {
                    "std": station_board_service.std,
                    "etd": station_board_service.etd,
                    "platform": station_board_service.platform,
                    "destination_location_name": station_board_service.destination.location[
                        0
                    ].locationName,
                    "origin_location_name": station_board_service.origin.location[
                        0
                    ].locationName,
                    "operator": station_board_service.operator,
                    "calling_points": [
                        {"location_name": cp.locationName, "st": cp.st, "et": cp.et}
                        for cp in station_board_service.subsequentCallingPoints.callingPointList[
                            0
                        ].callingPoint
                    ],
                }
            )
    except Exception:  # pylint: disable=broad-except
        return []

    return services


def next_services(station_code: str) -> List[Any]:
    # check parameters
    check_params(station_code)

    # get trains departing from station
    station_board_with_details = api.get_departure_board_with_details(station_code)
    services = services_from_station_board_with_details(station_board_with_details)

    return services



import pathlib
import json


DATA_DIRNAME = "data"
DATA_FILENAME = "referentiel-gares-voyageurs.json"


def _stations(data_dirname, data_filename):
    data_path = pathlib.Path(__file__).parents[0] / data_dirname / data_filename

    # read from stations data file
    with open(data_path) as data_file:
        raw_stations = json.load(data_file)

    stations = {}

    for raw_station in raw_stations:
        station_id = str(raw_station["fields"]["pltf_uic_code"])
        station_data = {"nom": raw_station["fields"]["gare_alias_libelle_noncontraint"]}
        stations[station_id] = station_data

    return stations


STATIONS = _stations(DATA_DIRNAME, DATA_FILENAME)



from . import data, commons, api


def check_params(stop_area_id: str = None):
    if not any((stop_area_id)):
        return

    stations = data.STATIONS

    if stop_area_id is not None:
        try:
            _ = stations[stop_area_id]
        except Exception as e:
            raise commons.SncfException(f"invalid stop area id {stop_area_id}") from e


def stations_by_string(string):
    string = str(string).lower()
    results = {}
    stations = data.STATIONS

    # iterate over stations
    for station_id in stations:
        # match?
        if string in stations[station_id]["nom"].lower():
            results[station_id] = stations[station_id]

    return results


def timetable_from_disruptions(disruptions):
    timetable = []

    impacted_stops = disruptions["disruptions"][0]["impacted_objects"][0][
        "impacted_stops"
    ]

    for impacted_stop in impacted_stops:
        # rewrite times from '162300' to '16:23'
        base_departure_time = impacted_stop["base_departure_time"]
        base_departure_time = f"{base_departure_time[:2]}:{base_departure_time[2:4]}"

        # determine (and rewrite) amended departure time
        if (impacted_stop["departure_status"] == "delayed") or (
            impacted_stop["departure_status"] == "unchanged"
        ):
            amended_departure_time = impacted_stop["amended_departure_time"]
            amended_departure_time = (
                f"{amended_departure_time[:2]}:{amended_departure_time[2:4]}"
            )
        elif impacted_stop["departure_status"] == "deleted":
            amended_departure_time = "--:--"
        else:  # 'added' - FIXME? may crash as never seen before
            amended_departure_time = "??:??"

        timetable.append(
            {
                "id": impacted_stop["stop_point"]["id"],
                "name": impacted_stop["stop_point"]["name"],
                "base_departure_time": base_departure_time,
                "amended_departure_time": amended_departure_time,
            }
        )

    return timetable


def timetable_from_vehicle_journeys(vehicle_journeys):
    timetable = []

    stop_times = vehicle_journeys["vehicle_journeys"][0]["stop_times"]

    for stop_time in stop_times:
        # rewrite time from '162300' to '16:23'
        departure_time = stop_time["departure_time"]
        departure_time = f"{departure_time[:2]}:{departure_time[2:4]}"

        timetable.append(
            {
                "id": stop_time["stop_point"]["id"],
                "name": stop_time["stop_point"]["name"],
                "base_departure_time": departure_time,
                "amended_departure_time": departure_time,
            }
        )

    return timetable


def timetable_from_departure(current_departure):
    timetable = []

    # schedule if disruption
    has_disruption = False
    for link in current_departure["display_informations"]["links"]:
        if link["type"] != "disruption":
            continue

        has_disruption = True
        disruptions = api.disruptions(link["id"])
        timetable = timetable_from_disruptions(disruptions)

        break

    if has_disruption and timetable != []:
        return timetable

    # schedule if normal traffic or disrupted schedule unresolved
    for link in current_departure["links"]:
        if link["type"] != "vehicle_journey":
            continue

        vehicle_journeys = api.vehicle_journeys(link["id"])
        timetable = timetable_from_vehicle_journeys(vehicle_journeys)

    return timetable


def next_trains(stop_area_id: str, timetable_for_all_trains: bool = False):
    # check parameters
    check_params(stop_area_id)

    # get trains departing from station
    departures = api.departures(stop_area_id)
    if departures is None:
        return []

    trains = []

    # extract data for each depature
    for i, current_departure in enumerate(departures["departures"]):
        # extract time for later rewrite (see below)
        departure_date_time = current_departure["stop_date_time"]["departure_date_time"]
        base_departure_date_time = current_departure["stop_date_time"][
            "base_departure_date_time"
        ]

        # remove city from direction, e.g. "Lyon-Part-Dieu (Lyon)" => "Lyon-Part-Dieu"
        direction = current_departure["display_informations"]["direction"]
        index_of_last_open_paren = direction.rfind("(")
        if index_of_last_open_paren != -1:
            direction = direction[: index_of_last_open_paren - 1]

        train_data = {
            "stop_point_id": current_departure["stop_point"]["id"],
            "direction": direction,
            "headsign": current_departure["display_informations"]["headsign"],
            # extract times from date (formatted as "20200822T203600")
            "base_time": f"{base_departure_date_time[-6:-4]}:{base_departure_date_time[-4:-2]}",
            "time": f"{departure_date_time[-6:-4]}:{departure_date_time[-4:-2]}",
            "commercial_mode": current_departure["display_informations"][
                "commercial_mode"
            ],
        }

        # get timetables for 2nd and following trains if requested
        if i == 0 or timetable_for_all_trains:
            train_data["timetable"] = timetable_from_departure(current_departure)

        trains.append(train_data)

    return trains



import threading
import time
import logging

from fastapi import APIRouter
from pydantic import BaseModel
from starlette.requests import Request

from . import ratp, commons, view_model, data_updater

logger = logging.getLogger(__name__)

router = APIRouter()


class StationDirection(BaseModel):
    line_id: str
    line_station_id: str
    direction: str


@router.get("/stations/{station_name_query}")
async def search_stations(station_name_query: str):
    return ratp.stations_by_name(station_name_query)


@router.get("/stations-line/{line_id}")
async def stations_line(line_id: str):
    return ratp.stations_by_line(line_id)


@router.get("/lines/{line_name_query}")
async def search_lines(line_name_query: str):
    return ratp.lines_by_name(line_name_query)


@router.get("/lines-by-code/{code_query}")
async def search_lines_by_code(code_query: str):
    return ratp.lines_by_code(code_query)


@router.get("/directions/{line_id}")
async def directions(line_id: str):
    return ratp.directions(line_id)


# debugging only
@router.get("/next/{line_id}/{line_station_id}/{direction}")
async def next_departures(line_id: str, line_station_id: str, direction: str):
    try:
        departures = ratp.next_departures(line_id, line_station_id, direction)
    except commons.RatpException as e:
        logger.warning(str(e))
        return {"status": "error", "message": str(e)}

    return {"status": "OK", "response": departures}


# curl -X POST http://localhost:8000/ratp/start-client \
#   -d {\"line_id\":\"RB\", \"line_station_id\":\"15\", \"direction\":\"A\"}
@router.post("/start-client")
async def start_client(station_direction: StationDirection, request: Request):
    # check params
    try:
        ratp.check_params(
            station_direction.line_id,
            station_direction.line_station_id,
            station_direction.direction,
        )
    except commons.RatpException as e:
        logger.warning(str(e))
        return {"status": "error", "message": str(e)}

    # stop board client if already running
    if request.app.board_client.running:
        request.app.board_client.running = False
        time.sleep(1)

    # start board
    threading.Thread(
        target=request.app.board_client.run,
        args=[
            view_model.ViewModelRatp_192_32_3_Rows_To_ProtocolBuffers,
            data_updater.DataUpdaterRatp,
            {
                "line_id": station_direction.line_id,
                "line_station_id": station_direction.line_station_id,
                "direction": station_direction.direction,
            },
        ],
    ).start()

    return {"status": "OK"}



import json
import os
import logging
from typing import List

import requests

from . import commons

logger = logging.getLogger(__name__)


def unified_api_request(base_url: str, base_queries: List[str] = None):
    # check environment variables
    commons.check_env_vars()

    # build URL and send request
    url = f'{base_url}?app_key={os.environ["TFL_APP_KEY"]}'
    if base_queries:
        url = f"{url}&{'&'.join(base_queries)}"

    try:
        response = requests.get(url, timeout=15)
    except requests.exceptions.Timeout:
        logger.warning("TfL unified API HTTP request timed out")
        return None
    except Exception as e:  # pylint: disable=broad-except
        logger.warning("TfL unified API HTTP request error: %s", str(e))
        return None

    # process response
    if response.status_code == 200:
        return json.loads(response.content.decode("utf-8"))

    return None


def line_stoppoints(line_id):
    base_url = f"https://api.tfl.gov.uk/Line/{line_id}/StopPoints"
    stoppoints = unified_api_request(base_url)

    return stoppoints


def line_arrivals(
    line_ids: List[str], station_id: str = ""  # default: all stations on lines
):
    base_url = f"https://api.tfl.gov.uk/Line/{','.join(line_ids)}/Arrivals/{station_id}"
    return unified_api_request(base_url)


def tube_arrivals(count: int = 2):
    base_url = "https://api.tfl.gov.uk/Mode/tube/Arrivals"
    base_queries = [f"count={count}"]

    return unified_api_request(base_url, base_queries)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: departure/board/departure.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="departure/board/departure.proto",
    package="departure.proto",
    syntax="proto3",
    serialized_options=None,
    serialized_pb=_b(
        '\n\x1f\x64\x65parture/board/departure.proto\x12\x0f\x64\x65parture.proto">\n\x05Pixel\x12\t\n\x01x\x18\x01 \x01(\x05\x12\t\n\x01y\x18\x02 \x01(\x05\x12\t\n\x01r\x18\x03 \x01(\x05\x12\t\n\x01g\x18\x04 \x01(\x05\x12\t\n\x01\x62\x18\x05 \x01(\x05"\x87\x01\n\x13\x42oardSectionContent\x12&\n\x06pixels\x18\x01 \x03(\x0b\x32\x16.departure.proto.Pixel\x12\x11\n\tcontent_w\x18\x02 \x01(\x05\x12\x11\n\tcontent_h\x18\x03 \x01(\x05\x12\x10\n\x08repeat_x\x18\x04 \x01(\x08\x12\x10\n\x08repeat_y\x18\x05 \x01(\x08":\n\nNoMovement\x12\x15\n\rx_offset_init\x18\x01 \x01(\x05\x12\x15\n\ry_offset_init\x18\x02 \x01(\x05"U\n\rStaticContent\x12\x16\n\x0etotal_duration\x18\x01 \x01(\x05\x12\x15\n\rx_offset_init\x18\x02 \x01(\x05\x12\x15\n\ry_offset_init\x18\x03 \x01(\x05"\xa0\x01\n\x10ScrollingContent\x12\x15\n\rstep_duration\x18\x01 \x01(\x05\x12\x13\n\x0btotal_steps\x18\x02 \x01(\x05\x12\x18\n\x10\x64\x65lta_x_per_step\x18\x03 \x01(\x05\x12\x18\n\x10\x64\x65lta_y_per_step\x18\x04 \x01(\x05\x12\x15\n\rx_offset_init\x18\x05 \x01(\x05\x12\x15\n\ry_offset_init\x18\x06 \x01(\x05"\xc4\x01\n\x08Movement\x12\x32\n\x0bno_movement\x18\x01 \x01(\x0b\x32\x1b.departure.proto.NoMovementH\x00\x12\x38\n\x0estatic_content\x18\x02 \x01(\x0b\x32\x1e.departure.proto.StaticContentH\x00\x12>\n\x11scrolling_content\x18\x03 \x01(\x0b\x32!.departure.proto.ScrollingContentH\x00\x42\n\n\x08movement"\xcd\x01\n\x1b\x42oardSectionOperationStatus\x12\x15\n\rsection_index\x18\x01 \x01(\x05\x12O\n\x06status\x18\x02 \x01(\x0e\x32?.departure.proto.BoardSectionOperationStatus.BoardSectionStatus"F\n\x12\x42oardSectionStatus\x12\x06\n\x02OK\x10\x00\x12\x14\n\x10SECTION_NOEXISTS\x10\x01\x12\x12\n\x0eSECTION_EXISTS\x10\x02"\xb1\x01\n\x19\x42oardSectionUpdateRequest\x12\x15\n\rsection_index\x18\x01 \x01(\x05\x12\x35\n\x07\x63ontent\x18\x02 \x01(\x0b\x32$.departure.proto.BoardSectionContent\x12+\n\x08movement\x18\x03 \x03(\x0b\x32\x19.departure.proto.Movement\x12\x19\n\x11\x63ontinue_movement\x18\x04 \x01(\x08"Z\n\x1a\x42oardSectionsUpdateRequest\x12<\n\x08requests\x18\x01 \x03(\x0b\x32*.departure.proto.BoardSectionUpdateRequest"[\n\x1b\x42oardSectionsUpdateResponse\x12<\n\x06status\x18\x01 \x03(\x0b\x32,.departure.proto.BoardSectionOperationStatus"z\n\x19\x42oardSectionCreateRequest\x12\x15\n\rsection_index\x18\x01 \x01(\x05\x12\x10\n\x08output_x\x18\x02 \x01(\x05\x12\x10\n\x08output_y\x18\x03 \x01(\x05\x12\x10\n\x08output_w\x18\x04 \x01(\x05\x12\x10\n\x08output_h\x18\x05 \x01(\x05"Z\n\x1a\x42oardSectionsCreateRequest\x12<\n\x08requests\x18\x01 \x03(\x0b\x32*.departure.proto.BoardSectionCreateRequest"[\n\x1b\x42oardSectionsCreateResponse\x12<\n\x06status\x18\x01 \x03(\x0b\x32,.departure.proto.BoardSectionOperationStatus"2\n\x19\x42oardSectionDeleteRequest\x12\x15\n\rsection_index\x18\x01 \x01(\x05"Z\n\x1a\x42oardSectionsDeleteRequest\x12<\n\x08requests\x18\x01 \x03(\x0b\x32*.departure.proto.BoardSectionDeleteRequest"[\n\x1b\x42oardSectionsDeleteResponse\x12<\n\x06status\x18\x01 \x03(\x0b\x32,.departure.proto.BoardSectionOperationStatus2\x80\x01\n\x0c\x42oardManager\x12p\n\x13\x42oardSectionsUpdate\x12+.departure.proto.BoardSectionsUpdateRequest\x1a,.departure.proto.BoardSectionsUpdateResponseb\x06proto3'
    ),
)


_BOARDSECTIONOPERATIONSTATUS_BOARDSECTIONSTATUS = _descriptor.EnumDescriptor(
    name="BoardSectionStatus",
    full_name="departure.proto.BoardSectionOperationStatus.BoardSectionStatus",
    filename=None,
    file=DESCRIPTOR,
    values=[
        _descriptor.EnumValueDescriptor(
            name="OK", index=0, number=0, serialized_options=None, type=None
        ),
        _descriptor.EnumValueDescriptor(
            name="SECTION_NOEXISTS",
            index=1,
            number=1,
            serialized_options=None,
            type=None,
        ),
        _descriptor.EnumValueDescriptor(
            name="SECTION_EXISTS", index=2, number=2, serialized_options=None, type=None
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=899,
    serialized_end=969,
)
_sym_db.RegisterEnumDescriptor(_BOARDSECTIONOPERATIONSTATUS_BOARDSECTIONSTATUS)


_PIXEL = _descriptor.Descriptor(
    name="Pixel",
    full_name="departure.proto.Pixel",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="x",
            full_name="departure.proto.Pixel.x",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="y",
            full_name="departure.proto.Pixel.y",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="r",
            full_name="departure.proto.Pixel.r",
            index=2,
            number=3,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="g",
            full_name="departure.proto.Pixel.g",
            index=3,
            number=4,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="b",
            full_name="departure.proto.Pixel.b",
            index=4,
            number=5,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=52,
    serialized_end=114,
)


_BOARDSECTIONCONTENT = _descriptor.Descriptor(
    name="BoardSectionContent",
    full_name="departure.proto.BoardSectionContent",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="pixels",
            full_name="departure.proto.BoardSectionContent.pixels",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="content_w",
            full_name="departure.proto.BoardSectionContent.content_w",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="content_h",
            full_name="departure.proto.BoardSectionContent.content_h",
            index=2,
            number=3,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="repeat_x",
            full_name="departure.proto.BoardSectionContent.repeat_x",
            index=3,
            number=4,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="repeat_y",
            full_name="departure.proto.BoardSectionContent.repeat_y",
            index=4,
            number=5,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=117,
    serialized_end=252,
)


_NOMOVEMENT = _descriptor.Descriptor(
    name="NoMovement",
    full_name="departure.proto.NoMovement",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="x_offset_init",
            full_name="departure.proto.NoMovement.x_offset_init",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="y_offset_init",
            full_name="departure.proto.NoMovement.y_offset_init",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=254,
    serialized_end=312,
)


_STATICCONTENT = _descriptor.Descriptor(
    name="StaticContent",
    full_name="departure.proto.StaticContent",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="total_duration",
            full_name="departure.proto.StaticContent.total_duration",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="x_offset_init",
            full_name="departure.proto.StaticContent.x_offset_init",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="y_offset_init",
            full_name="departure.proto.StaticContent.y_offset_init",
            index=2,
            number=3,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=314,
    serialized_end=399,
)


_SCROLLINGCONTENT = _descriptor.Descriptor(
    name="ScrollingContent",
    full_name="departure.proto.ScrollingContent",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="step_duration",
            full_name="departure.proto.ScrollingContent.step_duration",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="total_steps",
            full_name="departure.proto.ScrollingContent.total_steps",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="delta_x_per_step",
            full_name="departure.proto.ScrollingContent.delta_x_per_step",
            index=2,
            number=3,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="delta_y_per_step",
            full_name="departure.proto.ScrollingContent.delta_y_per_step",
            index=3,
            number=4,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="x_offset_init",
            full_name="departure.proto.ScrollingContent.x_offset_init",
            index=4,
            number=5,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="y_offset_init",
            full_name="departure.proto.ScrollingContent.y_offset_init",
            index=5,
            number=6,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=402,
    serialized_end=562,
)


_MOVEMENT = _descriptor.Descriptor(
    name="Movement",
    full_name="departure.proto.Movement",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="no_movement",
            full_name="departure.proto.Movement.no_movement",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="static_content",
            full_name="departure.proto.Movement.static_content",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="scrolling_content",
            full_name="departure.proto.Movement.scrolling_content",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="movement",
            full_name="departure.proto.Movement.movement",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=565,
    serialized_end=761,
)


_BOARDSECTIONOPERATIONSTATUS = _descriptor.Descriptor(
    name="BoardSectionOperationStatus",
    full_name="departure.proto.BoardSectionOperationStatus",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="section_index",
            full_name="departure.proto.BoardSectionOperationStatus.section_index",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="status",
            full_name="departure.proto.BoardSectionOperationStatus.status",
            index=1,
            number=2,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[
        _BOARDSECTIONOPERATIONSTATUS_BOARDSECTIONSTATUS,
    ],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=764,
    serialized_end=969,
)


_BOARDSECTIONUPDATEREQUEST = _descriptor.Descriptor(
    name="BoardSectionUpdateRequest",
    full_name="departure.proto.BoardSectionUpdateRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="section_index",
            full_name="departure.proto.BoardSectionUpdateRequest.section_index",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="content",
            full_name="departure.proto.BoardSectionUpdateRequest.content",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="movement",
            full_name="departure.proto.BoardSectionUpdateRequest.movement",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="continue_movement",
            full_name="departure.proto.BoardSectionUpdateRequest.continue_movement",
            index=3,
            number=4,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=972,
    serialized_end=1149,
)


_BOARDSECTIONSUPDATEREQUEST = _descriptor.Descriptor(
    name="BoardSectionsUpdateRequest",
    full_name="departure.proto.BoardSectionsUpdateRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="requests",
            full_name="departure.proto.BoardSectionsUpdateRequest.requests",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1151,
    serialized_end=1241,
)


_BOARDSECTIONSUPDATERESPONSE = _descriptor.Descriptor(
    name="BoardSectionsUpdateResponse",
    full_name="departure.proto.BoardSectionsUpdateResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="status",
            full_name="departure.proto.BoardSectionsUpdateResponse.status",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1243,
    serialized_end=1334,
)


_BOARDSECTIONCREATEREQUEST = _descriptor.Descriptor(
    name="BoardSectionCreateRequest",
    full_name="departure.proto.BoardSectionCreateRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="section_index",
            full_name="departure.proto.BoardSectionCreateRequest.section_index",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="output_x",
            full_name="departure.proto.BoardSectionCreateRequest.output_x",
            index=1,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="output_y",
            full_name="departure.proto.BoardSectionCreateRequest.output_y",
            index=2,
            number=3,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="output_w",
            full_name="departure.proto.BoardSectionCreateRequest.output_w",
            index=3,
            number=4,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
        _descriptor.FieldDescriptor(
            name="output_h",
            full_name="departure.proto.BoardSectionCreateRequest.output_h",
            index=4,
            number=5,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1336,
    serialized_end=1458,
)


_BOARDSECTIONSCREATEREQUEST = _descriptor.Descriptor(
    name="BoardSectionsCreateRequest",
    full_name="departure.proto.BoardSectionsCreateRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="requests",
            full_name="departure.proto.BoardSectionsCreateRequest.requests",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1460,
    serialized_end=1550,
)


_BOARDSECTIONSCREATERESPONSE = _descriptor.Descriptor(
    name="BoardSectionsCreateResponse",
    full_name="departure.proto.BoardSectionsCreateResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="status",
            full_name="departure.proto.BoardSectionsCreateResponse.status",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1552,
    serialized_end=1643,
)


_BOARDSECTIONDELETEREQUEST = _descriptor.Descriptor(
    name="BoardSectionDeleteRequest",
    full_name="departure.proto.BoardSectionDeleteRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="section_index",
            full_name="departure.proto.BoardSectionDeleteRequest.section_index",
            index=0,
            number=1,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1645,
    serialized_end=1695,
)


_BOARDSECTIONSDELETEREQUEST = _descriptor.Descriptor(
    name="BoardSectionsDeleteRequest",
    full_name="departure.proto.BoardSectionsDeleteRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="requests",
            full_name="departure.proto.BoardSectionsDeleteRequest.requests",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1697,
    serialized_end=1787,
)


_BOARDSECTIONSDELETERESPONSE = _descriptor.Descriptor(
    name="BoardSectionsDeleteResponse",
    full_name="departure.proto.BoardSectionsDeleteResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="status",
            full_name="departure.proto.BoardSectionsDeleteResponse.status",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1789,
    serialized_end=1880,
)

_BOARDSECTIONCONTENT.fields_by_name["pixels"].message_type = _PIXEL
_MOVEMENT.fields_by_name["no_movement"].message_type = _NOMOVEMENT
_MOVEMENT.fields_by_name["static_content"].message_type = _STATICCONTENT
_MOVEMENT.fields_by_name["scrolling_content"].message_type = _SCROLLINGCONTENT
_MOVEMENT.oneofs_by_name["movement"].fields.append(
    _MOVEMENT.fields_by_name["no_movement"]
)
_MOVEMENT.fields_by_name["no_movement"].containing_oneof = _MOVEMENT.oneofs_by_name[
    "movement"
]
_MOVEMENT.oneofs_by_name["movement"].fields.append(
    _MOVEMENT.fields_by_name["static_content"]
)
_MOVEMENT.fields_by_name["static_content"].containing_oneof = _MOVEMENT.oneofs_by_name[
    "movement"
]
_MOVEMENT.oneofs_by_name["movement"].fields.append(
    _MOVEMENT.fields_by_name["scrolling_content"]
)
_MOVEMENT.fields_by_name[
    "scrolling_content"
].containing_oneof = _MOVEMENT.oneofs_by_name["movement"]
_BOARDSECTIONOPERATIONSTATUS.fields_by_name[
    "status"
].enum_type = _BOARDSECTIONOPERATIONSTATUS_BOARDSECTIONSTATUS
_BOARDSECTIONOPERATIONSTATUS_BOARDSECTIONSTATUS.containing_type = (
    _BOARDSECTIONOPERATIONSTATUS
)
_BOARDSECTIONUPDATEREQUEST.fields_by_name["content"].message_type = _BOARDSECTIONCONTENT
_BOARDSECTIONUPDATEREQUEST.fields_by_name["movement"].message_type = _MOVEMENT
_BOARDSECTIONSUPDATEREQUEST.fields_by_name[
    "requests"
].message_type = _BOARDSECTIONUPDATEREQUEST
_BOARDSECTIONSUPDATERESPONSE.fields_by_name[
    "status"
].message_type = _BOARDSECTIONOPERATIONSTATUS
_BOARDSECTIONSCREATEREQUEST.fields_by_name[
    "requests"
].message_type = _BOARDSECTIONCREATEREQUEST
_BOARDSECTIONSCREATERESPONSE.fields_by_name[
    "status"
].message_type = _BOARDSECTIONOPERATIONSTATUS
_BOARDSECTIONSDELETEREQUEST.fields_by_name[
    "requests"
].message_type = _BOARDSECTIONDELETEREQUEST
_BOARDSECTIONSDELETERESPONSE.fields_by_name[
    "status"
].message_type = _BOARDSECTIONOPERATIONSTATUS
DESCRIPTOR.message_types_by_name["Pixel"] = _PIXEL
DESCRIPTOR.message_types_by_name["BoardSectionContent"] = _BOARDSECTIONCONTENT
DESCRIPTOR.message_types_by_name["NoMovement"] = _NOMOVEMENT
DESCRIPTOR.message_types_by_name["StaticContent"] = _STATICCONTENT
DESCRIPTOR.message_types_by_name["ScrollingContent"] = _SCROLLINGCONTENT
DESCRIPTOR.message_types_by_name["Movement"] = _MOVEMENT
DESCRIPTOR.message_types_by_name[
    "BoardSectionOperationStatus"
] = _BOARDSECTIONOPERATIONSTATUS
DESCRIPTOR.message_types_by_name[
    "BoardSectionUpdateRequest"
] = _BOARDSECTIONUPDATEREQUEST
DESCRIPTOR.message_types_by_name[
    "BoardSectionsUpdateRequest"
] = _BOARDSECTIONSUPDATEREQUEST
DESCRIPTOR.message_types_by_name[
    "BoardSectionsUpdateResponse"
] = _BOARDSECTIONSUPDATERESPONSE
DESCRIPTOR.message_types_by_name[
    "BoardSectionCreateRequest"
] = _BOARDSECTIONCREATEREQUEST
DESCRIPTOR.message_types_by_name[
    "BoardSectionsCreateRequest"
] = _BOARDSECTIONSCREATEREQUEST
DESCRIPTOR.message_types_by_name[
    "BoardSectionsCreateResponse"
] = _BOARDSECTIONSCREATERESPONSE
DESCRIPTOR.message_types_by_name[
    "BoardSectionDeleteRequest"
] = _BOARDSECTIONDELETEREQUEST
DESCRIPTOR.message_types_by_name[
    "BoardSectionsDeleteRequest"
] = _BOARDSECTIONSDELETEREQUEST
DESCRIPTOR.message_types_by_name[
    "BoardSectionsDeleteResponse"
] = _BOARDSECTIONSDELETERESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Pixel = _reflection.GeneratedProtocolMessageType(
    "Pixel",
    (_message.Message,),
    dict(
        DESCRIPTOR=_PIXEL,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.Pixel)
    ),
)
_sym_db.RegisterMessage(Pixel)

BoardSectionContent = _reflection.GeneratedProtocolMessageType(
    "BoardSectionContent",
    (_message.Message,),
    dict(
        DESCRIPTOR=_BOARDSECTIONCONTENT,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.BoardSectionContent)
    ),
)
_sym_db.RegisterMessage(BoardSectionContent)

NoMovement = _reflection.GeneratedProtocolMessageType(
    "NoMovement",
    (_message.Message,),
    dict(
        DESCRIPTOR=_NOMOVEMENT,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.NoMovement)
    ),
)
_sym_db.RegisterMessage(NoMovement)

StaticContent = _reflection.GeneratedProtocolMessageType(
    "StaticContent",
    (_message.Message,),
    dict(
        DESCRIPTOR=_STATICCONTENT,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.StaticContent)
    ),
)
_sym_db.RegisterMessage(StaticContent)

ScrollingContent = _reflection.GeneratedProtocolMessageType(
    "ScrollingContent",
    (_message.Message,),
    dict(
        DESCRIPTOR=_SCROLLINGCONTENT,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.ScrollingContent)
    ),
)
_sym_db.RegisterMessage(ScrollingContent)

Movement = _reflection.GeneratedProtocolMessageType(
    "Movement",
    (_message.Message,),
    dict(
        DESCRIPTOR=_MOVEMENT,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.Movement)
    ),
)
_sym_db.RegisterMessage(Movement)

BoardSectionOperationStatus = _reflection.GeneratedProtocolMessageType(
    "BoardSectionOperationStatus",
    (_message.Message,),
    dict(
        DESCRIPTOR=_BOARDSECTIONOPERATIONSTATUS,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.BoardSectionOperationStatus)
    ),
)
_sym_db.RegisterMessage(BoardSectionOperationStatus)

BoardSectionUpdateRequest = _reflection.GeneratedProtocolMessageType(
    "BoardSectionUpdateRequest",
    (_message.Message,),
    dict(
        DESCRIPTOR=_BOARDSECTIONUPDATEREQUEST,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.BoardSectionUpdateRequest)
    ),
)
_sym_db.RegisterMessage(BoardSectionUpdateRequest)

BoardSectionsUpdateRequest = _reflection.GeneratedProtocolMessageType(
    "BoardSectionsUpdateRequest",
    (_message.Message,),
    dict(
        DESCRIPTOR=_BOARDSECTIONSUPDATEREQUEST,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.BoardSectionsUpdateRequest)
    ),
)
_sym_db.RegisterMessage(BoardSectionsUpdateRequest)

BoardSectionsUpdateResponse = _reflection.GeneratedProtocolMessageType(
    "BoardSectionsUpdateResponse",
    (_message.Message,),
    dict(
        DESCRIPTOR=_BOARDSECTIONSUPDATERESPONSE,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.BoardSectionsUpdateResponse)
    ),
)
_sym_db.RegisterMessage(BoardSectionsUpdateResponse)

BoardSectionCreateRequest = _reflection.GeneratedProtocolMessageType(
    "BoardSectionCreateRequest",
    (_message.Message,),
    dict(
        DESCRIPTOR=_BOARDSECTIONCREATEREQUEST,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.BoardSectionCreateRequest)
    ),
)
_sym_db.RegisterMessage(BoardSectionCreateRequest)

BoardSectionsCreateRequest = _reflection.GeneratedProtocolMessageType(
    "BoardSectionsCreateRequest",
    (_message.Message,),
    dict(
        DESCRIPTOR=_BOARDSECTIONSCREATEREQUEST,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.BoardSectionsCreateRequest)
    ),
)
_sym_db.RegisterMessage(BoardSectionsCreateRequest)

BoardSectionsCreateResponse = _reflection.GeneratedProtocolMessageType(
    "BoardSectionsCreateResponse",
    (_message.Message,),
    dict(
        DESCRIPTOR=_BOARDSECTIONSCREATERESPONSE,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.BoardSectionsCreateResponse)
    ),
)
_sym_db.RegisterMessage(BoardSectionsCreateResponse)

BoardSectionDeleteRequest = _reflection.GeneratedProtocolMessageType(
    "BoardSectionDeleteRequest",
    (_message.Message,),
    dict(
        DESCRIPTOR=_BOARDSECTIONDELETEREQUEST,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.BoardSectionDeleteRequest)
    ),
)
_sym_db.RegisterMessage(BoardSectionDeleteRequest)

BoardSectionsDeleteRequest = _reflection.GeneratedProtocolMessageType(
    "BoardSectionsDeleteRequest",
    (_message.Message,),
    dict(
        DESCRIPTOR=_BOARDSECTIONSDELETEREQUEST,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.BoardSectionsDeleteRequest)
    ),
)
_sym_db.RegisterMessage(BoardSectionsDeleteRequest)

BoardSectionsDeleteResponse = _reflection.GeneratedProtocolMessageType(
    "BoardSectionsDeleteResponse",
    (_message.Message,),
    dict(
        DESCRIPTOR=_BOARDSECTIONSDELETERESPONSE,
        __module__="departure.board.departure_pb2"
        # @@protoc_insertion_point(class_scope:departure.proto.BoardSectionsDeleteResponse)
    ),
)
_sym_db.RegisterMessage(BoardSectionsDeleteResponse)


_BOARDMANAGER = _descriptor.ServiceDescriptor(
    name="BoardManager",
    full_name="departure.proto.BoardManager",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    serialized_start=1883,
    serialized_end=2011,
    methods=[
        _descriptor.MethodDescriptor(
            name="BoardSectionsUpdate",
            full_name="departure.proto.BoardManager.BoardSectionsUpdate",
            index=0,
            containing_service=None,
            input_type=_BOARDSECTIONSUPDATEREQUEST,
            output_type=_BOARDSECTIONSUPDATERESPONSE,
            serialized_options=None,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_BOARDMANAGER)

DESCRIPTOR.services_by_name["BoardManager"] = _BOARDMANAGER

# @@protoc_insertion_point(module_scope)



import json
import pathlib

import departure.provider.tfl_tube.tfl_tube as tfl_tube

# pylint: disable=too-many-lines


class TestIsDepartingFromStationInDirection:
    # {
    #     'naptanId': '940GZZLULGN',
    #     'lineId': 'central',
    #     'platformName': 'Platform 2/3',
    #     'direction': 'inbound',
    #     'vehicleId': '114',
    #     'currentLocation': 'At Platform',
    #     'towards': 'Northolt'
    # }

    def test_no_platform_direction_actual_direction(self):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "central",
            "940GZZLULGN",
            "Westbound",
            "940GZZLUNHT",
            platform_departure_direction=None,
            train_canonical_direction="inbound",
        )

    def test_no_platform_direction_incorrect_direction(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "central",
            "940GZZLULGN",
            "Eastbound",
            "940GZZLUNHT",
            platform_departure_direction=None,
            train_canonical_direction="inbound",
        )

    def test_no_platform_direction_equivalent_direction(self):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "central",
            "940GZZLULGN",
            "Outer Rail",
            "940GZZLUNHT",
            platform_departure_direction=None,
            train_canonical_direction="inbound",
        )

    def test_no_platform_direction_equivalent_incorrect_direction(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "central",
            "940GZZLULGN",
            "Inner Rail",
            "940GZZLUNHT",
            platform_departure_direction=None,
            train_canonical_direction="inbound",
        )

    def test_hainault_loop_actual_south_of_hlt(self):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "central",
            "940GZZLUFLP",
            "Westbound",
            None,
            platform_departure_direction="Outer Rail",
            train_canonical_direction="inbound",
        )

    def test_hainault_loop_actual_south_of_hlt_no_direction(self):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "central",
            "940GZZLUFLP",
            "Westbound",
            None,
            platform_departure_direction="Outer Rail",
            train_canonical_direction=None,
        )

    def test_hainault_loop_actual_north_of_hlt(self):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "central",
            "940GZZLURVY",
            "Westbound",
            None,
            platform_departure_direction="Outer Rail",
            train_canonical_direction="outbound",
        )

    def test_hainault_loop_actual_north_of_hlt_no_canonical_dir(self):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "central",
            "940GZZLURVY",
            "Westbound",
            None,
            platform_departure_direction="Outer Rail",
        )

    def test_cannot_determine(self):
        assert (
            tfl_tube.will_arriving_train_depart_from_station_in_direction(
                "district",
                "940GZZLUVIC",
                "Westbound",
                None,
                platform_departure_direction=None,
                train_canonical_direction=None,
            )
            is None
        )

    def test_at_terminating_station_no_departure_in_arrival_direction(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "district",
            "940GZZLUWIM",
            "Westbound",
            None,
            platform_departure_direction="Westbound",
            train_canonical_direction=None,
        )

    def test_at_terminating_station_no_departure_in_departure_direction(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "district",
            "940GZZLUWIM",
            "Westbound",
            None,
            platform_departure_direction="Eastbound",
            train_canonical_direction=None,
        )

    def test_at_inline_terminating_station_not_terminating_arr_direction(self):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "jubilee",
            "940GZZLUWYP",
            "Eastbound",
            None,
            platform_departure_direction="Eastbound",
            train_canonical_direction=None,
            destination_station_id="940GZZLUSTD",
        )

    def test_at_inline_terminating_station_not_terminating_dep_direction(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "jubilee",
            "940GZZLUWYP",
            "Westbound",
            None,
            platform_departure_direction="Eastbound",
            train_canonical_direction=None,
            destination_station_id="940GZZLUSTD",
        )

    def test_at_inline_terminating_station_not_terminating_arr_can_dir(self):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "jubilee",
            "940GZZLUWYP",
            "Eastbound",
            None,
            platform_departure_direction="Eastbound",
            train_canonical_direction="outbound",
            destination_station_id="940GZZLUSTD",
        )

    def test_at_inline_terminating_station_not_terminating_dep_can_dir(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "jubilee",
            "940GZZLUWYP",
            "Westbound",
            None,
            platform_departure_direction="Eastbound",
            train_canonical_direction="outbound",
            destination_station_id="940GZZLUSTD",
        )

    def test_at_inline_terminating_station_has_destination_station_id(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "jubilee",
            "940GZZLUWYP",
            "Westbound",
            None,
            platform_departure_direction="Eastbound",
            train_canonical_direction="outbound",
            destination_station_id="940GZZLUWYP",
        )

    def test_at_inline_terminating_station_has_towards_station_id(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "jubilee",
            "940GZZLUWYP",
            "Westbound",
            "940GZZLUWYP",
            platform_departure_direction="Eastbound",
            train_canonical_direction="outbound",
        )

    def test_at_inline_terminating_station_no_destination_dep_direction(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "jubilee",
            "940GZZLUWYP",
            "Westbound",
            None,
            platform_departure_direction="Eastbound",
            train_canonical_direction="outbound",
        )

    def test_at_inline_terminating_station_no_destination_arr_direction(self):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "jubilee",
            "940GZZLUWYP",
            "Eastbound",
            None,
            platform_departure_direction="Eastbound",
            train_canonical_direction="outbound",
        )

    def test_mismatching_direction_canonical_direction(self, caplog):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "district",
            "940GZZLUEMB",
            "Eastbound",
            None,
            platform_departure_direction="Eastbound",
            train_canonical_direction="inbound",
        )
        assert caplog.text.endswith(
            "incoherent direction at 940GZZLUEMB "
            "(district): towards None / destination None / Eastbound "
            "platform / inbound\n"
        )

    def test_circle_terminating_at_erc_arr_platform_arr_direction(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "circle",
            "940GZZLUERC",
            "Eastbound",
            "940GZZLUERC",
            platform_departure_direction="Eastbound",
        )

    def test_circle_terminating_at_erc_arr_platform_dep_direction(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "circle",
            "940GZZLUERC",
            "Westbound",
            "940GZZLUERC",
            platform_departure_direction="Eastbound",
        )

    def test_circle_terminating_at_erc_dep_platform_arr_direction(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "circle",
            "940GZZLUERC",
            "Eastbound",
            "940GZZLUERC",
            platform_departure_direction="Westbound",
        )

    def test_circle_terminating_at_erc_dep_platform_dep_direction(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "circle",
            "940GZZLUERC",
            "Westbound",
            "940GZZLUERC",
            platform_departure_direction="Westbound",
        )

    def test_circle_no_platform_south(self):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "circle", "940GZZLUWSM", "Westbound", "940GZZLUERC"
        )

    def test_circle_no_platform_north(self):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "circle", "940GZZLUBST", "Eastbound", "940GZZLUERC"
        )

    def test_circle_towards_erc_at_erc(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "circle", "940GZZLUERC", "Eastbound", "940GZZLUERC"
        )

    def test_circle_no_platform_no_destination_no_towards_at_erc(self):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "circle", "940GZZLUERC", "Eastbound", None
        )

    def test_circle_platform_no_destination_no_towards_at_erc_dep_dir(self):
        # not considered to be a terminus
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "circle",
            "940GZZLUERC",
            "Westbound",
            None,
            platform_departure_direction="Eastbound",
        )

    def test_circle_platform_no_destination_no_towards_at_erc_arr_dir(self):
        # not considered to be a terminus
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "circle",
            "940GZZLUERC",
            "Eastbound",
            None,
            platform_departure_direction="Eastbound",
        )

    def test_wrong_direction_prefer_platform_westbound(self, caplog):
        assert not tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "district",
            "940GZZLUWSM",
            "Westbound",
            "940GZZLUTWH",
            platform_departure_direction="Eastbound",
            train_canonical_direction="inbound",  # incoherent
        )
        assert caplog.text.endswith(
            "incoherent direction at 940GZZLUWSM "
            "(district): towards None / destination 940GZZLUTWH / Eastbound "
            "platform / inbound\n"
        )

    def test_wrong_direction_prefer_platform_eastbound(self, caplog):
        assert tfl_tube.will_arriving_train_depart_from_station_in_direction(
            "district",
            "940GZZLUWSM",
            "Eastbound",
            "940GZZLUTWH",
            platform_departure_direction="Eastbound",
            train_canonical_direction="inbound",  # incoherent
        )
        assert caplog.text.endswith(
            "incoherent direction at 940GZZLUWSM "
            "(district): towards None / destination 940GZZLUTWH / Eastbound "
            "platform / inbound\n"
        )


class TestIsStationTerminusForTrain:
    def test_destination_equals_station(self):
        assert tfl_tube.is_station_terminus_for_train(
            "circle", "940GZZLUERC", destination_station_id="940GZZLUERC"
        )

    def test_towards_equals_station(self):
        assert tfl_tube.is_station_terminus_for_train(
            "circle", "940GZZLUERC", towards_station_id="940GZZLUERC"
        )

    def test_terminus_end_of_line(self):
        assert tfl_tube.is_station_terminus_for_train("district", "940GZZLUWIM")


class TestDirectionFromPlatformName:
    def test_westbound(self):
        assert (
            tfl_tube.departure_direction_from_platform_name("Westbound - Platform 1")
            == "Westbound"
        )

    def test_eastbound(self):
        assert (
            tfl_tube.departure_direction_from_platform_name("Eastbound - Platform 1")
            == "Eastbound"
        )

    def test_northbound(self):
        assert (
            tfl_tube.departure_direction_from_platform_name("Northbound - Platform 1")
            == "Northbound"
        )

    def test_southbound(self):
        assert (
            tfl_tube.departure_direction_from_platform_name("Southbound - Platform 4")
            == "Southbound"
        )

    def test_northbound_fast(self):  # e.g. HOH
        assert (
            tfl_tube.departure_direction_from_platform_name(
                "Northbound Fast - Platform 3"
            )
            == "Northbound"
        )

    def test_southbound_fast(self):  # e.g. HOH
        assert (
            tfl_tube.departure_direction_from_platform_name(
                "Southbound Fast - Platform 6"
            )
            == "Southbound"
        )

    def test_outer(self):  # e.g. FLP
        assert (
            tfl_tube.departure_direction_from_platform_name(
                "Outer Rail - Platform 1",
            )
            == "Outer Rail"
        )

    def test_inner(self):  # e.g. FLP
        assert (
            tfl_tube.departure_direction_from_platform_name(
                "Inner Rail - Platform 2",
            )
            == "Inner Rail"
        )

    def test_westbound_no_dash(self):  # at WLO
        assert (
            tfl_tube.departure_direction_from_platform_name("Westbound Platform 26")
            == "Westbound"
        )

    def test_westbound_upper_b(self):  # at BDS
        assert (
            tfl_tube.departure_direction_from_platform_name("WestBound - Platform 2")
            == "Westbound"
        )

    def test_eastbound_upper_b(self):  # at BDS
        assert (
            tfl_tube.departure_direction_from_platform_name("EastBound - Platform 1")
            == "Eastbound"
        )

    def test_unmatched(self):  # at ASG
        assert (
            tfl_tube.departure_direction_from_platform_name(
                "Eastbound/Westbound - Platform 2/3"
            )
            is None
        )


class TestNumberFromPlatformName:
    def test_single(self):
        assert tfl_tube.number_from_platform_name("Westbound - Platform 1") == "1"

    def test_multiple(self):  # at ASG
        assert (
            tfl_tube.number_from_platform_name("Eastbound/Westbound - Platform 2/3")
            == "2/3"
        )

    def test_unmatched(self):  # at CSM
        assert tfl_tube.number_from_platform_name("North / South") is None


class TestStationIdFromTowards:
    def test_station_id_from_towards_normal(self):
        assert tfl_tube.station_id_from_towards("Hainault", "central") == "940GZZLUHLT"

    def test_station_id_from_towards_alt_name_ampersand(self):
        assert (
            tfl_tube.station_id_from_towards("Elephant and Castle", "bakerloo")
            == "940GZZLUEAC"
        )

    def test_station_id_from_towards_apostrophe(self):
        assert (
            tfl_tube.station_id_from_towards("Queen's Park", "bakerloo")
            == "940GZZLUQPS"
        )

    def test_station_id_from_towards_via_lowercase_v(self):
        assert (
            tfl_tube.station_id_from_towards("Hainault via Newbury Park", "central")
            == "940GZZLUHLT"
        )

    def test_station_id_from_towards_via_uppercase_v(self):
        assert (
            tfl_tube.station_id_from_towards("Woodford Via Hainault", "central")
            == "940GZZLUWOF"
        )

    def test_station_id_from_towards_alt_name_circle(self):
        assert (
            tfl_tube.station_id_from_towards("Edgware Road (Circle)", "circle")
            == "940GZZLUERC"
        )

    def test_station_id_from_towards_alt_name_hr5(self):
        assert (
            tfl_tube.station_id_from_towards("Heathrow T123 + 5", "piccadilly")
            == "940GZZLUHR5"
        )

    def test_station_id_from_towards_alt_name_hr4(self):
        assert (
            tfl_tube.station_id_from_towards("Heathrow via T4 Loop", "piccadilly")
            == "940GZZLUHR4"
        )

    def test_station_id_from_towards_not_a_station(self):
        assert (
            tfl_tube.station_id_from_towards("Check Front of Train", "circle") is None
        )


class TestLocationPlatformFromCurrentLocation:
    # sample data from:
    # jq -c "[.[] | select(.currentLocation | startswith(\"At \")) | \
    #    {currentLocation: .currentLocation, lineId: .lineId}] | unique" \
    #    sample-tube-arrivals-20181230T121143Z.json

    def test_location_platform(self):
        location, platform = tfl_tube.location_platform_from_current_location(
            "At Arsenal Platform 2"
        )
        assert (location, platform) == ("Arsenal", "2")

    def test_location_platform_two_digit_platform(self):
        location, platform = tfl_tube.location_platform_from_current_location(
            "At Baker Street Platform 10"
        )
        assert (location, platform) == ("Baker Street", "10")

    def test_location_platform_comma(self):
        location, platform = tfl_tube.location_platform_from_current_location(
            "At East Finchley, Platform 1"
        )
        assert (location, platform) == ("East Finchley", "1")

    def test_location_platform_typo_plaform(self):
        location, platform = tfl_tube.location_platform_from_current_location(
            "At London Bridge Plaform 1"
        )
        assert (location, platform) == ("London Bridge", "1")

    def test_location_platform_p_for_platform(self):
        location, platform = tfl_tube.location_platform_from_current_location(
            "At Kings Cross P7"
        )
        assert (location, platform) == ("Kings Cross", "7")

    def test_location_platform_no_number(self):
        location, platform = tfl_tube.location_platform_from_current_location(
            "At Angel platform"
        )
        assert (location, platform) == ("Angel", None)

    def test_location_no_platform(self):
        location, platform = tfl_tube.location_platform_from_current_location(
            "At Mornington Crescent"
        )
        assert (location, platform) == ("Mornington Crescent", None)

    def test_location_not_a_station(self):
        location, platform = tfl_tube.location_platform_from_current_location(
            "At North Acton Junction"
        )
        assert (location, platform) == ("North Acton Junction", None)

    def test_at_platform(self):
        location, platform = tfl_tube.location_platform_from_current_location(
            "At Platform"
        )
        assert (location, platform) == ("<current>", "<current>")

    def test_not_current_location_at(self):
        location, platform = tfl_tube.location_platform_from_current_location(
            "Not At East Finchley, Platform 1"
        )
        assert (location, platform) == (None, None)


class TestCurrentStationPlatformFromCurrentLocation:
    # sample data from:
    # jq "[.[] | select(.currentLocation | startswith(\"At \")) | \
    #     [.currentLocation, .lineId, .naptanId, .platformName]] | unique" \
    #     sample-tube-arrivals-20181230T121143Z.json

    def test_location_platform(self):
        (
            station,
            platform,
        ) = tfl_tube.current_station_platform_number_from_current_location(
            "At Arsenal Platform 2",
            "piccadilly",
            "940GZZLULSQ",
            "Westbound - Platform 1",
        )
        assert (station, platform) == ("940GZZLUASL", "2")

    def test_location_platform_two_digit_platform(self):
        (
            station,
            platform,
        ) = tfl_tube.current_station_platform_number_from_current_location(
            "At Baker Street Platform 10",
            "jubilee",
            "940GZZLUWHP",
            "Northbound - Platform 1",
        )
        assert (station, platform) == ("940GZZLUBST", "10")

    def test_location_platform_comma(self):
        (
            station,
            platform,
        ) = tfl_tube.current_station_platform_number_from_current_location(
            "At East Finchley, Platform 1",
            "northern",
            "940GZZLUWOP",
            "Northbound - Platform 1",
        )
        assert (station, platform) == ("940GZZLUEFY", "1")

    def test_location_platform_no_number(self):
        (
            station,
            platform,
        ) = tfl_tube.current_station_platform_number_from_current_location(
            "At Angel platform", "northern", "940GZZLUBTC", "Southbound - Platform 2"
        )
        assert (station, platform) == ("940GZZLUAGL", None)

    def test_location_no_platform(self):
        (
            station,
            platform,
        ) = tfl_tube.current_station_platform_number_from_current_location(
            "At Mornington Crescent",
            "northern",
            "940GZZLUWFN",
            "Northbound - Platform 1",
        )
        assert (station, platform) == ("940GZZLUMTC", None)

    def test_location_not_a_station(self):
        (
            station,
            platform,
        ) = tfl_tube.current_station_platform_number_from_current_location(
            "At North Acton Junction",
            "central",
            "940GZZLUTCR",
            "Eastbound - Platform 2",
        )
        assert (station, platform) == (None, None)

    def test_at_platform(self):
        (
            station,
            platform,
        ) = tfl_tube.current_station_platform_number_from_current_location(
            "At Platform",
            "piccadilly",
            "940GZZLUASG",
            "Eastbound/Westbound - Platform 2/3",
        )
        assert (station, platform) == ("940GZZLUASG", "2/3")

    def test_not_current_location_at(self):
        (
            station,
            platform,
        ) = tfl_tube.current_station_platform_number_from_current_location(
            "Not At East Finchley, Platform 1",
            "northern",
            "940GZZLUWOP",
            "Northbound - Platform 1",
        )
        assert (station, platform) == (None, None)


class TestStationFromLocation:
    def test_station(self):
        assert (
            tfl_tube.station_from_location("Gants Hill", None, "central")
            == "940GZZLUGTH"
        )

    def test_station_alternative_name(self):
        assert (
            tfl_tube.station_from_location("Elephant and Castle", None, "northern")
            == "940GZZLUEAC"
        )

    def test_not_station_known_location(self):
        assert (
            tfl_tube.station_from_location("North Acton Junction", None, "central")
            is None
        )

    def test_not_station_unknown_location(self, caplog):
        assert tfl_tube.station_from_location("Somewhere", None, "central") is None
        assert caplog.text.endswith("unresolved location Somewhere\n")


class TestCircleLineRailArrivingAtStation:
    def test_platform_westbound_outer_rail(self):
        assert (
            tfl_tube.circle_line_rail_arriving_at_station(
                "940GZZLUVIC", platform_departure_direction="Westbound"
            )
            == "Outer Rail"
        )

    def test_platform_westbound_inner_rail(self):
        assert (
            tfl_tube.circle_line_rail_arriving_at_station(
                "940GZZLUBST", platform_departure_direction="Westbound"
            )
            == "Inner Rail"
        )

    def test_platform_eastbound_outer_rail(self):
        assert (
            tfl_tube.circle_line_rail_arriving_at_station(
                "940GZZLUGPS", platform_departure_direction="Eastbound"
            )
            == "Outer Rail"
        )

    def test_platform_eastbound_outer_rail_erc(self):
        assert (
            tfl_tube.circle_line_rail_arriving_at_station(
                "940GZZLUERC", platform_departure_direction="Eastbound"
            )
            == "Outer Rail"
        )

    def test_platform_eastbound_inner_rail(self):
        assert (
            tfl_tube.circle_line_rail_arriving_at_station(
                "940GZZLUEMB", platform_departure_direction="Eastbound"
            )
            == "Inner Rail"
        )

    def test_terminating_erc(self):
        assert (
            tfl_tube.circle_line_rail_arriving_at_station(
                "940GZZLUMMT", destination_station_id="940GZZLUERC"
            )
            == "Outer Rail"
        )

    def test_terminating_hsc(self):
        assert (
            tfl_tube.circle_line_rail_arriving_at_station(
                "940GZZLUMMT", destination_station_id="940GZZLUHSC"
            )
            == "Inner Rail"
        )

    def test_towards_erc(self):
        assert (
            tfl_tube.circle_line_rail_arriving_at_station(
                "940GZZLUMMT", towards_station_id="940GZZLUERC"
            )
            == "Outer Rail"
        )

    def test_towards_hsc(self):
        assert (
            tfl_tube.circle_line_rail_arriving_at_station(
                "940GZZLUMMT", towards_station_id="940GZZLUHSC"
            )
            == "Inner Rail"
        )

    def test_undefined(self):
        assert tfl_tube.circle_line_rail_arriving_at_station("940GZZLUMMT") is None


class TestFilterTrainsByDirectionOfDeparture:
    def test_bkf_eastbound(self, caplog):
        """
        jq "[.[] | select(.naptanId == \"940GZZLUBKF\")]" \
            sample-line-district_circle-arrivals-20190111.json \
            > line-arrivals-circle_district-BKF.json
        """

        with open(
            pathlib.Path(__file__).parent
            / "data/line-arrivals-circle_district-BKF.json",
            "r",
        ) as file_handler:
            line_arrivals = json.load(file_handler)

        trains = tfl_tube.filter_trains_by_direction_of_departure(
            line_arrivals, "Eastbound"
        )

        assert len(trains) == 13

        assert trains["district-105"] == {
            "line_id": "district",
            "vehicle_id": "105",
            "station_id": "940GZZLUBKF",
            "towards": "Upminster",
            "towards_station_id": "940GZZLUUPM",
            "platform_name": "Eastbound - Platform 2",
            "platform_number": "2",
            "destination_station_id": "940GZZLUUPM",
            "time_to_station": 60,
        }

        assert trains["circle-213"] == {
            "line_id": "circle",
            "vehicle_id": "213",
            "station_id": "940GZZLUBKF",
            "towards": "Hammersmith",
            "towards_station_id": "940GZZLUHSC",
            "platform_name": "Eastbound - Platform 2",
            "platform_number": "2",
            "destination_station_id": "940GZZLUHSC",
            "time_to_station": 361,
        }

        assert trains["district-031"] == {
            "line_id": "district",
            "vehicle_id": "031",
            "station_id": "940GZZLUBKF",
            "towards": "Tower Hill",
            "towards_station_id": "940GZZLUTWH",
            "platform_name": "Eastbound - Platform 2",
            "platform_number": "2",
            "destination_station_id": "940GZZLUTWH",
            "time_to_station": 480,
        }

        assert trains["district-055"]
        assert trains["district-003"]
        assert trains["circle-214"]
        assert trains["district-030"]
        assert trains["district-064"]
        assert trains["district-032"]
        assert trains["district-107"]
        assert trains["district-042"]
        assert trains["district-005"]

        # fake entry with incoherent direction (inbound) vs platform
        # (Eastbound); actual is Eastbound
        assert trains["district-063"]
        assert caplog.text.endswith(
            "incoherent direction at 940GZZLUWSM "
            "(district): towards 940GZZLUTWH / destination 940GZZLUTWH / "
            "Eastbound platform / inbound\n"
        )

    def test_bkf_westbound(self, caplog):
        with open(
            pathlib.Path(__file__).parent
            / "data/line-arrivals-circle_district-BKF.json",
            "r",
        ) as file_handler:
            line_arrivals = json.load(file_handler)

        trains = tfl_tube.filter_trains_by_direction_of_departure(
            line_arrivals, "Westbound"
        )

        assert len(trains) == 12

        assert trains["district-017"] == {
            "line_id": "district",
            "vehicle_id": "017",
            "station_id": "940GZZLUBKF",
            "towards": "Wimbledon",
            "towards_station_id": "940GZZLUWIM",
            "platform_name": "Westbound - Platform 1",
            "platform_number": "1",
            "destination_station_id": "940GZZLUWIM",
            "time_to_station": 91,
        }

        assert trains["district-040"] == {
            "line_id": "district",
            "vehicle_id": "040",
            "station_id": "940GZZLUBKF",
            "towards": "Richmond",
            "towards_station_id": "940GZZLURMD",
            "platform_name": "Westbound - Platform 1",
            "platform_number": "1",
            "destination_station_id": "940GZZLURMD",
            "time_to_station": 211,
        }

        assert trains["district-125"] == {
            "line_id": "district",
            "vehicle_id": "125",
            "station_id": "940GZZLUBKF",
            "towards": "Wimbledon",
            "towards_station_id": "940GZZLUWIM",
            "platform_name": "Westbound - Platform 1",
            "platform_number": "1",
            "destination_station_id": "940GZZLUWIM",
            "time_to_station": 271,
        }

        assert trains["circle-202"] == {
            "line_id": "circle",
            "vehicle_id": "202",
            "station_id": "940GZZLUBKF",
            "towards": "Edgware Road (Circle)",
            "towards_station_id": "940GZZLUERC",
            "platform_name": "Westbound - Platform 1",
            "platform_number": "1",
            "destination_station_id": "940GZZLUERC",
            "time_to_station": 480,
        }

        assert trains["district-014"]
        assert trains["district-016"]
        assert trains["district-126"]
        assert trains["circle-203"]
        assert trains["district-052"]
        assert trains["district-056"]
        assert trains["district-060"]
        assert trains["circle-204"]

        # fake entry with incoherent direction (inbound) vs platform
        # (Eastbound); actual is Eastbound
        assert caplog.text.endswith(
            "incoherent direction at 940GZZLUWSM "
            "(district): towards 940GZZLUTWH / destination 940GZZLUTWH / "
            "Eastbound platform / inbound\n"
        )

    def test_bst_eastbound(self):
        """
        jq "[.[] | select(.naptanId == \"940GZZLUBST\")]" \
            sample-line-district_circle-arrivals-20190111.json \
            > line-arrivals-circle_district-BST.json
        """

        with open(
            pathlib.Path(__file__).parent
            / "data/line-arrivals-circle_district-BST.json",
            "r",
        ) as file_handler:
            line_arrivals = json.load(file_handler)

        trains = tfl_tube.filter_trains_by_direction_of_departure(
            line_arrivals, "Eastbound"
        )

        assert len(trains) == 1

        assert trains["circle-201"] == {
            "line_id": "circle",
            "vehicle_id": "201",
            "station_id": "940GZZLUBST",
            "towards": "Edgware Road (Circle)",
            "towards_station_id": "940GZZLUERC",
            "platform_name": "Eastbound - Platform 5",
            "platform_number": "5",
            "destination_station_id": "940GZZLUERC",
            "time_to_station": 1741,
        }

    def test_bst_westbound(self):
        with open(
            pathlib.Path(__file__).parent
            / "data/line-arrivals-circle_district-BST.json",
            "r",
        ) as file_handler:
            line_arrivals = json.load(file_handler)

        trains = tfl_tube.filter_trains_by_direction_of_departure(
            line_arrivals, "Westbound"
        )

        assert len(trains) == 0


class TestFilterTrainsCurrentlyAtPlatform:
    def test_bkf_circle_district_eastbound(self):
        """
        jq "[.[] | select(\
                (.currentLocation | startswith(\"At Blackfriars\")) or \
                (\
                    (.naptanId == \"940GZZLUBKF\") and \
                    (.currentLocation | startswith(\"At Platform\"))
                )\
            )]" sample-tube-arrivals-20190111.json > tube-arrivals-at_BKF.json
        """

        with open(
            pathlib.Path(__file__).parent / "data/tube-arrivals-at_BKF.json", "r"
        ) as file_handler:
            tube_arrivals = json.load(file_handler)

        trains = tfl_tube.filter_trains_currently_at_platform(
            tube_arrivals,
            line_ids=["district", "circle"],
            requested_station_id="940GZZLUBKF",
            requested_direction="Eastbound",
        )

        assert len(trains) == 1

        assert trains["district-106"] == {
            "line_id": "district",
            "vehicle_id": "106",
            "station_id": "940GZZLUBKF",
            "towards": "Upminster",
            "towards_station_id": "940GZZLUUPM",
            "platform_number": "2",
            "destination_station_id": "940GZZLUUPM",
            "time_to_station": 0,
        }

    def test_bkf_circle_district_westbound(self):
        with open(
            pathlib.Path(__file__).parent / "data/tube-arrivals-at_BKF.json", "r"
        ) as file_handler:
            tube_arrivals = json.load(file_handler)

        trains = tfl_tube.filter_trains_currently_at_platform(
            tube_arrivals,
            line_ids=["district", "circle"],
            requested_station_id="940GZZLUBKF",
            requested_direction="Westbound",
        )

        assert len(trains) == 1

        assert trains["circle-201"] == {
            "line_id": "circle",
            "vehicle_id": "201",
            "station_id": "940GZZLUBKF",
            "towards": "Edgware Road (Circle)",
            "towards_station_id": "940GZZLUERC",
            "platform_number": "1",
            "destination_station_id": "940GZZLUERC",
            "time_to_station": 0,
        }

    def test_bst_circle_district_eastbound(self):
        """
        jq "[.[] | select(\
                (.currentLocation | startswith(\"At Baker Street\")) or \
                (\
                    (.naptanId == \"940GZZLUBST\") and \
                    (.currentLocation | startswith(\"At Platform\"))
                )\
            )]" sample-tube-arrivals-20190111.json > tube-arrivals-at_BST.json
        """

        with open(
            pathlib.Path(__file__).parent / "data/tube-arrivals-at_BST.json", "r"
        ) as file_handler:
            tube_arrivals = json.load(file_handler)

        trains = tfl_tube.filter_trains_currently_at_platform(
            tube_arrivals,
            line_ids=["district", "circle"],
            requested_station_id="940GZZLUBST",
            requested_direction="Eastbound",
        )

        assert len(trains) == 1

        assert trains["circle-204"] == {
            "line_id": "circle",
            "vehicle_id": "204",
            "station_id": "940GZZLUBST",
            "towards": "Edgware Road (Circle)",
            "towards_station_id": "940GZZLUERC",
            "platform_number": "5",
            "destination_station_id": "940GZZLUERC",
            "time_to_station": 0,
        }

    def test_bst_circle_district_westbound(self):
        with open(
            pathlib.Path(__file__).parent / "data/tube-arrivals-at_BST.json", "r"
        ) as file_handler:
            tube_arrivals = json.load(file_handler)

        trains = tfl_tube.filter_trains_currently_at_platform(
            tube_arrivals,
            line_ids=["district", "circle"],
            requested_station_id="940GZZLUBST",
            requested_direction="Westbound",
        )

        assert len(trains) == 0


class TestWillArrivingTrainDepartFromCurrentLocationInDirection:
    def test_arrived_at_terminus_arrival_direction(self):
        assert (
            tfl_tube.will_arriving_train_depart_from_current_location_in_direction(
                "940GZZLUBNK",
                "waterloo-city",
                "940GZZLUBNK",
                "Westbound",
                "940GZZLUBNK",
            )
            == False
        )

    def test_arrived_at_terminus_departure_direction(self):
        assert (
            tfl_tube.will_arriving_train_depart_from_current_location_in_direction(
                "940GZZLUBNK",
                "waterloo-city",
                "940GZZLUBNK",
                "Eastbound",
                "940GZZLUBNK",
            )
            == False
        )

    def test_departing_from_terminus_towards(self):
        assert (
            tfl_tube.will_arriving_train_depart_from_current_location_in_direction(
                "940GZZLUEBY", "central", "940GZZLUEBY", "Eastbound", "940GZZLUHLT"
            )
            == True
        )

    def test_departing_from_terminus_departure_direction(self):
        assert (
            tfl_tube.will_arriving_train_depart_from_current_location_in_direction(
                "940GZZLUEBY", "central", "940GZZLUEBY", "Eastbound", "940GZZLUWCY"
            )
            == True
        )


class TestDirectionOfTrainFromStationTerminatingAt:
    def test_end_of_line(self):
        assert (
            tfl_tube.canonical_direction_of_train_from_station_terminating_at(
                "central", "940GZZLUEBY", "940GZZLUHLT"
            )
            == "outbound"
        )

    def test_inline_single_terminating_direction(self):
        assert (
            tfl_tube.canonical_direction_of_train_from_station_terminating_at(
                "district", "940GZZLUVIC", "940GZZLUTWH"
            )
            == "outbound"
        )

    def test_inline_dual_terminating_directions_normal_service(self):
        assert (
            tfl_tube.canonical_direction_of_train_from_station_terminating_at(
                "central", "940GZZLULYS", "940GZZLUWCY"
            )
            == "inbound"
        )

    def test_inline_dual_terminating_directions_end_of_service(self):
        assert (
            tfl_tube.canonical_direction_of_train_from_station_terminating_at(
                "central", "940GZZLUEBY", "940GZZLUWCY"
            )
            == "outbound"
        )


class TestResolveTrainCanonicalDirection:
    def test_towards_end_of_line(self):
        assert (
            tfl_tube.resolve_train_canonical_direction(
                "central", "940GZZLUOXC", towards_station_id="940GZZLUEBY"
            )
            == "inbound"
        )

    def test_towards_end_of_line_incoherent(self, caplog):
        assert (
            tfl_tube.resolve_train_canonical_direction(
                "central",
                "940GZZLUOXC",
                towards_station_id="940GZZLUEBY",
                destination_station_id="940GZZLUHLT",
            )
            == "inbound"
        )

        assert caplog.text.endswith(
            "incoherent direction at 940GZZLUOXC "
            "(central): towards 940GZZLUHLT / destination 940GZZLUEBY / "
            "None platform / None\n"
        )

    def test_destination_end_of_line(self):
        assert (
            tfl_tube.resolve_train_canonical_direction(
                "central", "940GZZLUOXC", destination_station_id="940GZZLUEBY"
            )
            == "inbound"
        )

    # TODO implement
    """
    def test_inline_single_terminating_direction(self):
        assert False

    def test_inline_single_terminating_direction_incoherent(self, caplog):
        assert False

    def test_inline_dual_term_dir_normal_service(self):
        assert False

    def test_inline_dual_term_dir_normal_service_incoherent(self, caplog):
        assert False

    def test_inline_dual_term_dir_end_of_service(self):
        assert False

    def test_inline_dual_term_dir_end_of_service_incoherent(self, caplog):
        assert False

    def test_from_platform(self):
        assert False

    def test_from_platform_incoherent(self, caplog):
        assert False

    def test_from_canonical_direction(self):
        assert False
    """


class TestInverseCanonicalDirection:
    def test_inbound(self):
        assert tfl_tube.inverse_canonical_direction("inbound") == "outbound"

    def test_outbound(self):
        assert tfl_tube.inverse_canonical_direction("outbound") == "inbound"

    def test_random(self):
        assert tfl_tube.inverse_canonical_direction("random") is None



from . import api, commons


def check_params(line_id: str, station_id: str, direction_sens: str):
    response = api.get_missions_next(line_id, station_id, direction_sens)

    # invalid station
    if response.ambiguityMessage is not None:
        raise commons.RatpException(
            "invalid line code and/or station code and/or direction"
        )


def lines_by_name(name_query: str = ""):
    results = []

    # get all lines
    lines = api.get_lines_realtime_realm()

    # filter by name
    for line in sorted(lines, key=lambda k: k["code"]):
        # ignore lines that do not contain query string
        if name_query not in line.name.lower():
            continue

        results.append(simplified_line_data(line))
    return results


def lines_by_code(code_query):
    results = []

    lines = api.get_lines_by_code(code_query)

    for line in sorted(lines, key=lambda k: k["code"]):
        # stop if no results
        if line["id"] is None:
            break

        results.append(simplified_line_data(line))

    return results


def simplified_line_data(line):
    if line.image is None:
        # image = 'data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7'
        image = ""
    else:
        image = f"http://opendata-tr.ratp.fr/wsiv/static/line/{line.image}"

    return {
        "id": line.id,
        "reseau": line.reseau.name,
        "code": line.code,
        "name": line.name,
        "image": image,
    }


def simplified_station_data(station: dict) -> dict:
    return {
        "name": station.name,
        "line_station_id": station.id,
        "line": simplified_line_data(station.line),
    }


def directions(line_id: str) -> dict:
    response = api.get_directions(line_id)

    results = {}
    for direction in response.directions:
        results[direction.sens] = direction.name

    return results


def stations_by_line(line_id: str) -> list:
    stations = api.get_stations_by_line(line_id)

    results = []
    for station in stations:
        results.append(simplified_station_data(station))

    return results


def stations_by_name(name_query: str) -> list:
    response = api.get_stations_by_name(name_query)

    results = []
    for station in response.stations:
        results.append(simplified_station_data(station))

    return results


def next_departures(line_id: str, station_id: str, direction_sens: str) -> dict:
    response = api.get_missions_next(line_id, station_id, direction_sens)

    # invalid station
    if response.ambiguityMessage is not None:
        raise commons.RatpException("invalid station code and/or direction")

    missions = []
    for mission in response.missions:
        # handle case where service has ended
        if len(mission.stations) > 1:
            destination_name = mission.stations[1].name
        else:
            destination_name = "---"

        if len(mission.stationsDates) > 0:
            stations_date = mission.stationsDates[0]
        else:
            stations_date = "------------"

        missions.append(
            {
                "code": mission.code,
                "destinationName": destination_name,
                "datetime": stations_date,
                # 'platform': mission.stationsPlatforms[0],  # only for RER
                "message": mission.stationsMessages[0],
            }
        )

    if len(response.perturbations) > 0:
        perturbations = response.perturbations[0].message.text
    else:
        perturbations = ""

    return {"missions": missions, "perturbations": perturbations}



import time
import threading
import logging

import departure.board.data_updater as data_updater
from departure.commons.log import log_function_stdout_to_debug
from . import view_model, tfl_tube, ui

logger = logging.getLogger(__name__)


class DataUpdaterTflTube(data_updater.DataUpdater):
    def __init__(
        self,
        target_view_model: view_model.ViewModelTflTube,
        data_refresh_delay_in_s: int,
        end_event: threading.Event,
        line_id: str,
        station_id: str,
        direction: str,
    ):
        super().__init__(target_view_model, data_refresh_delay_in_s, end_event)
        self.line_id = line_id
        self.station_id = station_id
        self.direction = direction
        self.trains = None

    def update(self):
        update_start_time = time.monotonic()
        logger.info("data refresh started at %s", update_start_time)

        trains = tfl_tube.next_trains(self.line_id, self.station_id, self.direction)

        # debugging: capture list of trains to string, then log
        log_function_stdout_to_debug(logger, ui.list_trains_single_station, trains)

        # update board only if there were changes
        if trains == self.trains:
            logger.info("no change - no update sent to board")
        else:
            self.view_model.update(trains)
            self.trains = trains

        logger.info("data refresh duration %s", time.monotonic() - update_start_time)



"""
Core functions for Nederlandse Spoorwegen
"""

from . import api, data, commons


def check_params(station_code: str):
    stations = data.STATIONS_BY_NS_CODE

    if station_code not in stations:
        raise commons.NSException(f"invalid station code {station_code}")


def departures_with_schedule(
    station_code: str, timetable_for_all_trains: bool = False
) -> list:
    trains = departures(station_code)

    for train in trains:
        train["schedule"] = stops(train["train_number"])
        if not timetable_for_all_trains:
            break  # then stop after first train

    return trains


def departures(station_code: str) -> list:
    check_params(station_code)

    departures_json = api.departures(station_code)

    trains = []

    for current_departure in departures_json["payload"]["departures"]:
        if "plannedTrack" in current_departure:
            track = current_departure["plannedTrack"]
        else:
            track = ""

        trains.append(
            {
                "direction": current_departure["direction"],
                "train_long_name": current_departure["name"],
                "planned_time": current_departure["plannedDateTime"][11:16],
                "actual_time": current_departure["actualDateTime"][11:16],
                "track": track,
                "train_number": current_departure["product"]["number"],
                "train_category": current_departure["product"]["longCategoryName"],
                "train_operator": current_departure["product"]["operatorName"],
            }
        )

    return trains


def stops(train_number: str) -> list:
    journey_json = api.journey(train_number)

    schedule = []
    for stop in journey_json["payload"]["stops"]:
        if "arrivals" in stop:
            direction = "arrivals"
        elif "departures" in stop:
            direction = "departures"
        else:
            continue

        planned_time = stop[direction][0]["plannedTime"][11:16]
        if "actualTime" in stop[direction][0]:
            actual_time = stop[direction][0]["actualTime"][11:16]
        else:
            actual_time = f"{planned_time}t"

        schedule.append(
            {
                "stop_name": stop["stop"]["name"],
                "stop_uic": stop["stop"]["uicCode"],
                "planned_time": planned_time,
                "actual_time": actual_time,
            }
        )

    return schedule


def search(station_name_query: str) -> dict:
    stations = data.STATIONS_BY_NS_CODE
    station_name_query = str(station_name_query).lower()

    results = {}

    for station_code in stations:
        if station_name_query in stations[station_code]["name_long"].lower():
            results[station_code] = stations[station_code]

    return results
